# Generated by the gRPC Python protocol compiler plugin. DO NOT EDIT!
"""Client and server classes corresponding to protobuf-defined services."""
import grpc
import warnings

from . import device_agent_pb2 as grpc__stubs_dot_device__agent__pb2

GRPC_GENERATED_VERSION = '1.65.1'
GRPC_VERSION = grpc.__version__
EXPECTED_ERROR_RELEASE = '1.66.0'
SCHEDULED_RELEASE_DATE = 'August 6, 2024'
_version_not_supported = False

try:
    from grpc._utilities import first_version_is_lower
    _version_not_supported = first_version_is_lower(GRPC_VERSION, GRPC_GENERATED_VERSION)
except ImportError:
    _version_not_supported = True

if _version_not_supported:
    warnings.warn(
        f'The grpc package installed is at version {GRPC_VERSION},'
        + f' but the generated code in grpc_stubs/device_agent_pb2_grpc.py depends on'
        + f' grpcio>={GRPC_GENERATED_VERSION}.'
        + f' Please upgrade your grpc module to grpcio>={GRPC_GENERATED_VERSION}'
        + f' or downgrade your generated code using grpcio-tools<={GRPC_VERSION}.'
        + f' This warning will become an error in {EXPECTED_ERROR_RELEASE},'
        + f' scheduled for release on {SCHEDULED_RELEASE_DATE}.',
        RuntimeWarning
    )


class deviceAgentStub(object):
    """The doover device agent service definition.
    """

    def __init__(self, channel):
        """Constructor.

        Args:
            channel: A grpc.Channel.
        """
        self.TestComms = channel.unary_unary(
                '/device_agent.deviceAgent/TestComms',
                request_serializer=grpc__stubs_dot_device__agent__pb2.TestCommsRequest.SerializeToString,
                response_deserializer=grpc__stubs_dot_device__agent__pb2.TestCommsResponse.FromString,
                _registered_method=True)
        self.GetChannelDetails = channel.unary_unary(
                '/device_agent.deviceAgent/GetChannelDetails',
                request_serializer=grpc__stubs_dot_device__agent__pb2.ChannelDetailsRequest.SerializeToString,
                response_deserializer=grpc__stubs_dot_device__agent__pb2.ChannelDetailsResponse.FromString,
                _registered_method=True)
        self.GetChannelSubscription = channel.unary_stream(
                '/device_agent.deviceAgent/GetChannelSubscription',
                request_serializer=grpc__stubs_dot_device__agent__pb2.ChannelSubscriptionRequest.SerializeToString,
                response_deserializer=grpc__stubs_dot_device__agent__pb2.ChannelSubscriptionResponse.FromString,
                _registered_method=True)
        self.WriteToChannel = channel.unary_unary(
                '/device_agent.deviceAgent/WriteToChannel',
                request_serializer=grpc__stubs_dot_device__agent__pb2.ChannelWriteRequest.SerializeToString,
                response_deserializer=grpc__stubs_dot_device__agent__pb2.ChannelWriteResponse.FromString,
                _registered_method=True)
        self.GetDebugInfo = channel.unary_unary(
                '/device_agent.deviceAgent/GetDebugInfo',
                request_serializer=grpc__stubs_dot_device__agent__pb2.DebugInfoRequest.SerializeToString,
                response_deserializer=grpc__stubs_dot_device__agent__pb2.DebugInfoResponse.FromString,
                _registered_method=True)
        self.GetTempAPIToken = channel.unary_unary(
                '/device_agent.deviceAgent/GetTempAPIToken',
                request_serializer=grpc__stubs_dot_device__agent__pb2.TempAPITokenRequest.SerializeToString,
                response_deserializer=grpc__stubs_dot_device__agent__pb2.TempAPITokenResponse.FromString,
                _registered_method=True)


class deviceAgentServicer(object):
    """The doover device agent service definition.
    """

    def TestComms(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def GetChannelDetails(self, request, context):
        """rpc GetAgentDetails (AgentDetailsRequest) returns (AgentDetailsResponse) {}
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def GetChannelSubscription(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def WriteToChannel(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def GetDebugInfo(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def GetTempAPIToken(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')


def add_deviceAgentServicer_to_server(servicer, server):
    rpc_method_handlers = {
            'TestComms': grpc.unary_unary_rpc_method_handler(
                    servicer.TestComms,
                    request_deserializer=grpc__stubs_dot_device__agent__pb2.TestCommsRequest.FromString,
                    response_serializer=grpc__stubs_dot_device__agent__pb2.TestCommsResponse.SerializeToString,
            ),
            'GetChannelDetails': grpc.unary_unary_rpc_method_handler(
                    servicer.GetChannelDetails,
                    request_deserializer=grpc__stubs_dot_device__agent__pb2.ChannelDetailsRequest.FromString,
                    response_serializer=grpc__stubs_dot_device__agent__pb2.ChannelDetailsResponse.SerializeToString,
            ),
            'GetChannelSubscription': grpc.unary_stream_rpc_method_handler(
                    servicer.GetChannelSubscription,
                    request_deserializer=grpc__stubs_dot_device__agent__pb2.ChannelSubscriptionRequest.FromString,
                    response_serializer=grpc__stubs_dot_device__agent__pb2.ChannelSubscriptionResponse.SerializeToString,
            ),
            'WriteToChannel': grpc.unary_unary_rpc_method_handler(
                    servicer.WriteToChannel,
                    request_deserializer=grpc__stubs_dot_device__agent__pb2.ChannelWriteRequest.FromString,
                    response_serializer=grpc__stubs_dot_device__agent__pb2.ChannelWriteResponse.SerializeToString,
            ),
            'GetDebugInfo': grpc.unary_unary_rpc_method_handler(
                    servicer.GetDebugInfo,
                    request_deserializer=grpc__stubs_dot_device__agent__pb2.DebugInfoRequest.FromString,
                    response_serializer=grpc__stubs_dot_device__agent__pb2.DebugInfoResponse.SerializeToString,
            ),
            'GetTempAPIToken': grpc.unary_unary_rpc_method_handler(
                    servicer.GetTempAPIToken,
                    request_deserializer=grpc__stubs_dot_device__agent__pb2.TempAPITokenRequest.FromString,
                    response_serializer=grpc__stubs_dot_device__agent__pb2.TempAPITokenResponse.SerializeToString,
            ),
    }
    generic_handler = grpc.method_handlers_generic_handler(
            'device_agent.deviceAgent', rpc_method_handlers)
    server.add_generic_rpc_handlers((generic_handler,))
    server.add_registered_method_handlers('device_agent.deviceAgent', rpc_method_handlers)


 # This class is part of an EXPERIMENTAL API.
class deviceAgent(object):
    """The doover device agent service definition.
    """

    @staticmethod
    def TestComms(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/device_agent.deviceAgent/TestComms',
            grpc__stubs_dot_device__agent__pb2.TestCommsRequest.SerializeToString,
            grpc__stubs_dot_device__agent__pb2.TestCommsResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def GetChannelDetails(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/device_agent.deviceAgent/GetChannelDetails',
            grpc__stubs_dot_device__agent__pb2.ChannelDetailsRequest.SerializeToString,
            grpc__stubs_dot_device__agent__pb2.ChannelDetailsResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def GetChannelSubscription(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_stream(
            request,
            target,
            '/device_agent.deviceAgent/GetChannelSubscription',
            grpc__stubs_dot_device__agent__pb2.ChannelSubscriptionRequest.SerializeToString,
            grpc__stubs_dot_device__agent__pb2.ChannelSubscriptionResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def WriteToChannel(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/device_agent.deviceAgent/WriteToChannel',
            grpc__stubs_dot_device__agent__pb2.ChannelWriteRequest.SerializeToString,
            grpc__stubs_dot_device__agent__pb2.ChannelWriteResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def GetDebugInfo(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/device_agent.deviceAgent/GetDebugInfo',
            grpc__stubs_dot_device__agent__pb2.DebugInfoRequest.SerializeToString,
            grpc__stubs_dot_device__agent__pb2.DebugInfoResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def GetTempAPIToken(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/device_agent.deviceAgent/GetTempAPIToken',
            grpc__stubs_dot_device__agent__pb2.TempAPITokenRequest.SerializeToString,
            grpc__stubs_dot_device__agent__pb2.TempAPITokenResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)



import copy
import enum
import inspect
import logging
import time
import json
from datetime import datetime

from typing import Union, Any, Optional, TypeVar

from .element import Element
from .interaction import SlimCommand, Interaction, NotSet
from .submodule import Container, NAME_VALIDATOR
from .variable import Variable

from ..cloud.api import Client

from .utils import find_object_with_key, find_path_to_key
# from ..docker.device_agent.device_agent import DeviceAgentInterface ## Causes circular import error
from ..utils import call_maybe_async, get_is_async, maybe_async

log = logging.getLogger(__name__)
ElementT = TypeVar("ElementT", bound=Element)
InteractionT = TypeVar("InteractionT", bound=Interaction)


class ShouldPushUpdate(enum.Enum):
    push_and_log = 1
    push_only = 2
    do_nothing = 3


class UIManager:
    def __init__(
        self,
        agent_id: str = None,
        client = None, # client: Union[Client, DeviceAgentInterface] = None, # Circular import
        auto_start: bool = False,
        min_ui_update_period: int = 600,
        min_observed_update_period: int = 4,
        is_async: bool = None,
    ):
        self._is_async = get_is_async(is_async)
        self.client = client
        # to determine whether we can use event-based logic. some reason we can't
        self._has_persistent_connection = client and client.has_persistent_connection()
        self._subscriptions_ready = False

        self.agent_id = agent_id

        self.last_ui_state = dict()  # A python dictionary of the full state from the cloud
        self.last_ui_state_update = None

        self.last_ui_state_wss_connections = dict()
        self.last_ui_state_wss_connections_update = None

        self.last_ui_cmds = dict()
        self.last_ui_cmds_update = None

        self._base_container = Container(name=None, display_name=None)
        self._interactions: dict[str, Interaction] = dict()

        self._has_critical_interaction_pending: bool = False
        # self._has_critical_ui_state_pending: bool = False
        self._critical_values = dict()  # legacy

        self.min_ui_update_period = min_ui_update_period
        self.min_observed_update_period = min_observed_update_period
        self._last_pushed_time = None
        self._last_push_and_log_time = None

        # legacy, list of subscriptions to call when we have a command update.
        self._cmds_subscriptions = []

        if auto_start:
            self.start_comms()

    def start_comms(self):
        self._setup_subscriptions()

    def _is_conn_ready(self, setup: bool = False) -> bool:
        if not self._has_persistent_connection:
            return self.client is not None

        if self._subscriptions_ready:
            return True
        elif setup:
            self._setup_subscriptions()
            return self._is_conn_ready(setup=False)  # don't setup for a second time
        else:
            log.error("Attempted use of dda_iface in ui_manager without dda_iface being ready")
            return False

    def is_connected(self) -> bool:
        if not self._has_persistent_connection:
            return self._is_conn_ready()

        if not self._is_conn_ready():
            return False

        return self.client.get_is_dda_online()

    get_is_connected = is_connected

    def is_being_observed(self):
        if not self.last_ui_state_wss_connections:
            return False

        try:
            if not isinstance(self.last_ui_state_wss_connections, dict):
                self.last_ui_state_wss_connections = json.loads(self.last_ui_state_wss_connections)
            connections = self.last_ui_state_wss_connections["connections"]
            # The following isn't working currently as agent_id is None
            # for k in connections.keys():
            #     if k != self.agent_id and connections[k] is True:
            #         return True

            # if there is more than one connection, then we are being observed
            return len(connections.keys()) > 1
        except Exception as e:
            log.error(f"Error checking if being observed: {e}")
            return False

    def has_been_connected(self):
        if self._has_persistent_connection and self.client is not None:
            return self.client.get_has_dda_been_online()
        return self.last_ui_state is not None

    get_has_been_connected = has_been_connected

    def _setup_subscriptions(self):
        if not self._has_persistent_connection:
            log.error("Attempted to setup subscriptions without valid connection client.")
            return

        if self.client is None:
            log.warning("Attempted to setup subscriptions without client being set")
            return

        log.info("Setting up dda subscriptions")
        self.client.add_subscription("ui_state", self.on_state_update)
        self.client.add_subscription("ui_state@wss_connections", self.on_state_wss_update)
        self.client.add_subscription("ui_cmds", self.on_command_update_async)

        self._subscriptions_ready = True

    async def await_comms_sync(self, timeout: int = 5) -> bool:
        if not self._has_persistent_connection:
            return False

        if not hasattr(self.client, "wait_for_channels_sync_async"):
            log.error("Attempted to await comms sync without valid connection client.")
            return False
        return await self.client.wait_for_channels_sync_async(
            channel_names=["ui_state", "ui_cmds"],
            timeout=timeout,
        )


    async def on_state_update(self, _, aggregate: dict[str, Any]):
        self._set_new_ui_state(aggregate)

    async def on_state_wss_update(self, _, aggregate: dict[str, Any]):
        self.last_ui_state_wss_connections = aggregate
        self.last_ui_state_wss_connections_update = time.time()

    def _on_command_update_common(self, aggregate: dict[str, Any]):
        aggregate = self._set_new_ui_cmds(aggregate)

        # add commands that don't currently exist
        to_add = {k: v for k, v in aggregate.items() if k not in self._interactions}
        for name, current_value in to_add.items():
            self._interactions[name] = SlimCommand(name, current_value)

        return aggregate

    def on_command_update(self, _, aggregate: dict[str, Any]):
        self._on_command_update_common(aggregate)

    async def on_command_update_async(self, _, aggregate: dict[str, Any]):
        prev_agg = copy.deepcopy(self.last_ui_cmds)
        aggregate = self._on_command_update_common(aggregate)
        # call all subscribed to cmds updates
        for c in self._cmds_subscriptions:
            await call_maybe_async(c)

        # work out command diff and call individual commands
        changed = {c: v for c, v in aggregate.items() if v != prev_agg.get(c)}
        for command_name, new_value in changed.items():
            command = self.get_command(command_name)
            if command is not None:
                await command._handle_new_value(new_value)

    def _set_new_ui_cmds(self, payload: dict[str, Any]):
        if not payload:
            log.info("Received empty UI commands payload.")
            payload = {}

        if isinstance(payload, str):
            try:
                payload = json.loads(payload)
            except json.JSONDecodeError:
                log.error(f"Failed to decode UI commands: {payload}")
                payload = {}

        try:
            payload = payload["cmds"]
        except KeyError:
            pass

        self.last_ui_cmds = payload
        self.last_ui_cmds_update = time.time()
        return payload

    def _set_new_ui_state(self, payload: dict[str, Any]):
        if not isinstance(payload, dict):
            payload = {}

        try:
            payload = payload["state"]
        except KeyError:
            pass

        self.last_ui_state = payload
        self.last_ui_state_update = time.time()

        ## TODO: Implement this ????
        # if self._base_container is not None:
        #     self._base_container.from_dict(payload)

        ## Iterate through the payload and update anything that needs updating
        # define a function to recursively trawl through each element of the last ui state and allow the element to update itself
        def update_elements_from_ui_state(d):
            for k, v in d.items():
                if isinstance(v, dict):
                    element = self.get_element(k)
                    if element is not None:
                        element.recv_ui_state_update(v)
                    else:
                        update_elements_from_ui_state(v)

        update_elements_from_ui_state(payload)


    def _add_interaction(self, interaction: Interaction):
        name = interaction.name.strip()
        if not NAME_VALIDATOR.match(name):
            raise RuntimeError(
                f"Invalid name '{name}' for interaction '{interaction}'. "
                f"Valid characters include letters, numbers, and underscores."
            )
        
        if name in self._interactions:
            ## If the interaction already exists, we should preserve the current value if it exists
            if hasattr(self._interactions[name], "_current_value") and self._interactions[name]._current_value is not NotSet:
                interaction.current_value = self._interactions[name]._current_value

        self._interactions[name] = interaction
        interaction._manager = self

    def _remove_interaction(self, interaction_name: str) -> None:
        try:
            del self._interactions[interaction_name]
        except KeyError:
            return

    def add_interaction(self, interaction: InteractionT):
        if not isinstance(interaction, Interaction) and hasattr(interaction, "_ui_type"):
            interaction = self._register_interaction(interaction, interaction.__self__)

        self._add_interaction(interaction)
        self._base_container.add_children(interaction)

    @staticmethod
    def _register_interaction(func, parent) -> InteractionT:
        item = func._ui_type(**func._ui_kwargs)
        item.callback = func

        try:
            setattr(parent, func.__name__, item)
        except AttributeError:
            pass  # maybe they've initialised the function w/o a class

        return item

    def register_interactions(self, obj_to_search):
        for name, func in inspect.getmembers(
            obj_to_search, predicate=lambda f: inspect.ismethod(f) and hasattr(f, "_ui_type")
        ):
            self._register_interaction(func, obj_to_search)

    def get_interaction(self, name: str) -> Optional[Interaction]:
        try:
            return self._interactions[name]
        except KeyError:
            return None
        
    def update_interaction(self, name: str, updated: Interaction) -> bool:
        if name not in self._interactions:
            return False
        
        self._interactions[name] = updated
        return True

    add_command = add_interaction
    get_command = get_interaction

    def coerce_command(self, command_name: str, value: Any, critical: bool = False) -> None:
        command = self.get_command(command_name)
        if command is None:
            log.info(f"Tried to coerce command {command_name} that doesn't exist.")
            return

        command.coerce(value, critical=critical)

    def get_element(self, element_name: str) -> Optional[ElementT]:
        result = self._base_container.get_element(element_name)
        # if not result:
        #     result = find_object_with_key(self.last_ui_state, element_name)
        return result
    
    def get_from_ui_state(self, element_name: str) -> Optional[dict]:
        return find_object_with_key(self.last_ui_state, element_name)

    def update_variable(self, variable_name: str, value: Any, critical: bool = False) -> bool:
        element = self._base_container.get_element(variable_name)
        if not (element and isinstance(element, Variable)):
            log.info(f"PyDoover: Tried to update variable '{variable_name}' that doesn't exist.")
            return False

        if critical is True and element.current_value != value:
            self._has_critical_interaction_pending = True

        element.current_value = value
        return True

    def add_cmds_update_subscription(self, callback):
        # fixme: create alias or something
        self._cmds_subscriptions.append(callback)

    def get_all_interactions(self) -> list[Interaction]:
        return list(self._interactions.values())

    def get_all_interaction_names(self) -> list[str]:
        return list(self._interactions.keys())

    get_available_commands = get_all_interaction_names

    def record_critical_value(self, name, value):
        log.warning("this function is deprecated. use the critical=True parameter of another appropriate function.")
        if self._critical_values.get(name) == value:
            return

        self._critical_values[name] = value
        self._has_critical_interaction_pending = True
        # self._has_critical_ui_state_pending = True

    def _should_push_update(self) -> ShouldPushUpdate:
        if self._has_critical_interaction_pending:
            return ShouldPushUpdate.push_and_log
        if self._last_pushed_time is None or self._last_push_and_log_time is None:
            return ShouldPushUpdate.push_and_log

        since_last_push = time.time() - self._last_pushed_time
        since_last_push_and_log = time.time() - self._last_push_and_log_time
        if since_last_push_and_log > self.min_ui_update_period:
            return ShouldPushUpdate.push_and_log
        elif self.is_being_observed() and since_last_push > self.min_observed_update_period:
            return ShouldPushUpdate.push_only

        return ShouldPushUpdate.do_nothing

    @maybe_async()
    def handle_comms(self, force_log: bool = False):
        should_push = self._should_push_update()

        if force_log is False and should_push is ShouldPushUpdate.do_nothing:
            return # don't need to push anything yet...
        
        self.push(record_log=force_log or should_push is ShouldPushUpdate.push_and_log)
        
    async def handle_comms_async(self, force_log: bool = False):
        should_push = self._should_push_update()

        if force_log is False and should_push is ShouldPushUpdate.do_nothing:
            return  # don't need to push anything yet...

        await self.push_async(record_log=force_log or should_push is ShouldPushUpdate.push_and_log)

    def _publish_to_channel(self, channel_name: str, data: dict[str, Any], record_log: bool = True, timestamp: Optional[datetime] = None, **kwargs):
        # this purely exists to provide cross-compatibility between clients (hence private method).
        if isinstance(self.client, Client):
            channel = self.client.get_channel_named(channel_name, self.agent_id)
            return channel.publish(data, save_log=record_log, timestamp=timestamp, **kwargs)
        else:
            # fixme: allow for timestamp in DDA message publishing...
            return self.client.publish_to_channel(channel_name, data, record_log=record_log, **kwargs)

    async def _publish_to_channel_async(self, channel_name: str, data: dict[str, Any], record_log: bool = True, timestamp: Optional[datetime] = None, **kwargs):
        if isinstance(self.client, Client):
            # in theory this works but lets just discourage this behaviour...
            raise RuntimeError("Cannot push async with a Client object")

        return await self.client.publish_to_channel_async(channel_name, data, record_log=record_log, **kwargs)

    @maybe_async()
    def pull(self):
        print("pulling...")
        if isinstance(self.client, Client):
            ui_cmds = self.client.get_channel_named("ui_cmds", self.agent_id)
            ui_state = self.client.get_channel_named("ui_state", self.agent_id)

            ui_cmds_agg = ui_cmds.fetch_aggregate()
            ui_state_agg = ui_state.fetch_aggregate()
        else:
            ui_cmds_agg = self.client.get_channel_aggregate("ui_cmds")
            ui_state_agg = self.client.get_channel_aggregate("ui_state")

        self._set_new_ui_state(ui_state_agg)
        
        # self._set_new_ui_cmds(ui_cmds_agg)
        self.on_command_update(None, ui_cmds_agg)

    async def pull_async(self):
        if isinstance(self.client, Client):
            raise RuntimeError("Cannot pull async with a Client object")

        ui_cmds_agg = await self.client.get_channel_aggregate_async("ui_cmds")
        ui_state_agg = await self.client.get_channel_aggregate_async("ui_state")
        self._set_new_ui_state(ui_state_agg)
        # self._set_new_ui_cmds(ui_cmds_agg)
        await self.on_command_update_async(None, ui_cmds_agg)

    def _check_dda_ready(self):
        if not self._is_conn_ready():
            log.warning("Attempted to push config without ready connection client.")
            return False
        elif not self.client.get_has_dda_been_online():
            # for a persistent connection, don't push if we haven't first pulled last data
            # HTTP-based connections will do a pull before pushing so that is fine.
            log.warning("Attempted to push config without DDA being online.")
            return False
        elif self.last_ui_state_update is None:
            log.warning("Waiting for UI state update to be pulled before pushing...")
            return False
        elif self.last_ui_cmds_update is None:
            log.warning("Waiting for UI commands to be pulled before pushing...")
            return False
        return True

    @maybe_async()
    def push(self,
            record_log: bool = True,
            should_remove: bool = True,
            timestamp: Optional[datetime] = None,
            even_if_empty: bool = False,
            only_channels: Optional[list] = None,
            publish_fields: Optional[list] = None,
        ) -> bool:
        publish_fields = publish_fields or []

        # self.check_dda()
        if self._has_persistent_connection:
            if not self._check_dda_ready():
                return False
        else:
            self.pull()  # do a pull before HTTP client pushes anything...

        print("pushing sync...")
        commands_update = self._get_commands_update(publish_fields=publish_fields)
        if commands_update is not None and (only_channels is None or "ui_cmds" in only_channels):
            self._publish_to_channel("ui_cmds", {"cmds": commands_update}, timestamp=timestamp)

        ui_state_update = self._get_ui_state_update(should_remove=should_remove, retain_fields=publish_fields)
        if even_if_empty or (ui_state_update is not None and (only_channels is None or "ui_state" in only_channels)):
            self._publish_to_channel("ui_state", ui_state_update or {}, record_log=record_log, timestamp=timestamp)
        else:
            print("not pushing empty ui state")

        self._last_pushed_time = time.time()
        if record_log:
            self._last_push_and_log_time = time.time()

        self._has_critical_interaction_pending = False
        return True

    async def push_async(self,
        record_log: bool = True,
        should_remove: bool = True,
        timestamp: Optional[datetime] = None,
        even_if_empty: bool = False,
        only_channels: Optional[list] = None,
        publish_fields: Optional[list] = None,
    ) -> bool:
        publish_fields = publish_fields or []

        # self.check_dda()
        if self._has_persistent_connection:
            if not self._check_dda_ready():
                return False
        else:
            self.pull()  # do a pull before HTTP client pushes anything...

        print("pushing async...")
        commands_update = self._get_commands_update(publish_fields=publish_fields)
        if commands_update is not None and (only_channels is None or "ui_cmds" in only_channels):
            await self._publish_to_channel_async("ui_cmds", {"cmds": commands_update}, timestamp=timestamp)

        ui_state_update = self._get_ui_state_update(should_remove=should_remove, retain_fields=publish_fields)
        if even_if_empty or (ui_state_update is not None and (only_channels is None or "ui_state" in only_channels)):
            await self._publish_to_channel_async("ui_state", ui_state_update or {}, record_log=record_log, timestamp=timestamp)
        else:
            print("not pushing empty ui state")

        self._last_pushed_time = time.time()
        if record_log:
            self._last_push_and_log_time = time.time()

        self._has_critical_interaction_pending = False
        return True

    @maybe_async()
    def clear_ui(self):
        # this could be dangerous...
        log.info("Clearing UI")
        self._publish_to_channel("ui_state", {"state": None})

    async def clear_ui_async(self):
        log.info("Clearing UI")
        await self._publish_to_channel_async("ui_state", {"state": None})

    def _get_commands_update(self, publish_fields: Optional[list] = None) -> Optional[dict[str, Any]]:
        if publish_fields is None:
            publish_fields = []

        cloud_commands = copy.deepcopy(self.last_ui_cmds)
        local_commands = {k: v._json_safe_current_value() for k, v in self._interactions.items()}

        # don't include commands that are the same as the cloud, and values that aren't set
        result = {
            name: value for name, value in local_commands.items()
            if cloud_commands.get(name) != value and value != NotSet
            or name in publish_fields
        }

        log.debug("Last Commands: " + str(cloud_commands))
        log.debug("New Commands: " + str(local_commands))
        log.debug("Commands Update: " + str(result))

        # don't clean up commands that exist upstream but not locally for now.
        result.update({c: None for c in cloud_commands.keys() if c not in local_commands})

        if len(result) == 0:
            return None
        return result

    def _get_ui_state_update(self, should_remove: bool = True, retain_fields: Optional[list] = []) -> Optional[dict[str, Any]]:
        cloud_state = self.last_ui_state or {}
        # this recursively evaluates and finds the diff on all children, rather than trying to do the diff here
        result = self._base_container.get_diff(cloud_state, remove=should_remove, retain_fields=retain_fields)

        log.debug("Last UI State: " + str(cloud_state))
        log.debug("New UI State: " + str(self._base_container.to_dict()))
        log.debug("UI State Update: " + str(result))

        if not result or len(result) == 0:
            return None

        return {"state": result}

    def _maybe_add_interaction_from_elems(self, *elements: Union[Element, Container]) -> list[Element]:
        to_return = []
        for element in elements:
            # this is a bit hacky, but it's to stop passing in an unregistered interaction (ie. created with a decorator
            # outside of a submodule and hasn't been registered yet),
            # instead we'll silently register it and proceed as-is
            if not isinstance(element, Interaction) and hasattr(element, "_ui_type"):
                element = self._register_interaction(element, element.__self__)

            if isinstance(element, Container):
                self._maybe_add_interaction_from_elems(*element.children)
            elif isinstance(element, Interaction):
                self._add_interaction(element)
            to_return.append(element)

        return to_return

    def add_children(self, *children: Element) -> None:
        if len(children) == 1 and isinstance(children[0], list):
            # for backwards compatibility, this used to accept a single list of children
            children = children[0]

        updated = self._maybe_add_interaction_from_elems(*children)
        self._base_container.add_children(*updated)

    def remove_children(self, *children: Element) -> None:
        if len(children) == 1 and isinstance(children[0], list):
            # for backwards compatibility, this used to accept a single list of children
            children = children[0]

        for elem in children:
            if not isinstance(elem, Element):
                # sometimes an unregistered function can end up here and break things...
                continue

            if elem == self._base_container:
                raise RuntimeError("You can't remove the base container!")

            # this should never be None, but in case some numpty does something weird...
            if getattr(elem, "parent", None):
                elem.parent.remove_children(elem)

            ## Remove the element
            self._base_container.remove_children(elem)
            self._remove_interaction(elem.name)

    def set_children(self, children: list[Element]) -> None:
        updated = self._maybe_add_interaction_from_elems(*children)
        self._base_container.set_children(updated)
        # self._maybe_add_interaction_from_elems(*children)
        # self._base_container.set_children(children)

        # self._base_container.add_children( self.cameras )
        # if len(self.cameras) > 0:
        #     self._base_container.add_children( [ doover_ui_hidden_value(name="last_cam_snapshot") ] )

    def set_status_icon(self, icon_type: str, critical: bool = False):
        if icon_type == self._base_container.status_icon:
            return
        elif critical is True:
            # self._has_critical_ui_state_pending = True
            self._has_critical_interaction_pending = True  # fixme: work out if we ever need this in element setting

        self._base_container.status_icon = icon_type

    def set_display_name(self, name: str, critical: bool = False):
        if name == self._base_container.display_name:
            return
        elif critical is True:
            self._has_critical_interaction_pending = True

        self._base_container.display_name = name

    set_display_str = set_display_name


ui_manager = UIManager



from typing import Union
from datetime import datetime

from .interaction import Interaction, NotSet


class Parameter(Interaction):
    type = NotImplemented


class NumericParameter(Parameter):
    type = "uiFloatParam"

    def __init__(
        self, name, display_name, min_val: Union[int, float] = None, max_val: Union[int, float] = None, **kwargs
    ):
        super().__init__(name, display_name, **kwargs)

        self.min = min_val or kwargs.pop("float_min", None)
        self.max = max_val or kwargs.pop("float_max", None)

    def to_dict(self):
        result = super().to_dict()

        if self.min is not None:
            result['min'] = self.min
        if self.max is not None:
            result['max'] = self.max

        return result


class TextParameter(Parameter):
    type = "uiTextParam"

    def __init__(self, name, display_name, is_text_area: bool = False, **kwargs):
        super().__init__(name, display_name, **kwargs)
        self.is_text_area = is_text_area

    def to_dict(self):
        result = super().to_dict()
        result['isTextArea'] = self.is_text_area
        return result


class BooleanParameter(Parameter):
    type = "uiBoolParam"

    def __init__(self, name, display_name, **kwargs):
        super().__init__(name, display_name, **kwargs)
        raise NotImplementedError("boolean parameter not implemented in doover site.")


class DateTimeParameter(Parameter):
    """Datetime is stored as epoch seconds UTC"""
    type = "uiDatetimeParam"

    def __init__(self, name: str, display_name: str, include_time: bool = False, **kwargs):
        super().__init__(name, display_name, **kwargs)
        self.include_time = include_time

    @property
    def current_value(self):
        if self._current_value is NotSet or self._current_value is None:
            return None
        if isinstance(self._current_value, datetime):
            return self._current_value
        elif isinstance(self._current_value, (int, float)):
            return datetime.utcfromtimestamp(self._current_value)
        return None
    
    @current_value.setter
    def current_value(self, new_val):
        if isinstance(new_val, datetime):
            new_val = int(new_val.timestamp())
        self._current_value = new_val

    def to_dict(self):
        result = super().to_dict()
        result['includeTime'] = self.include_time
        return result


doover_ui_float_parameter = NumericParameter
doover_ui_text_parameter = TextParameter
doover_ui_datetime_parameter = DateTimeParameter


def numeric_parameter(
    name: str, display_name: str, min_val: Union[int, float] = None, max_val: Union[int, float] = None, **kwargs
):
    def decorator(func):
        func._ui_type = NumericParameter
        func._ui_kwargs = {
            "name": name,
            "display_name": display_name,
            "min_val": min_val,
            "max_val": max_val,
            **kwargs,
        }
        return func
    return decorator

def text_parameter(name: str, display_name: str, is_text_area: bool = False, **kwargs):
    def decorator(func):
        func._ui_type = TextParameter
        func._ui_kwargs = {
            "name": name,
            "display_name": display_name,
            "is_text_area": is_text_area,
            **kwargs,
        }
        return func
    return decorator


def boolean_parameter(name: str, display_name: str, **kwargs):
    def decorator(func):
        func._ui_type = BooleanParameter
        func._ui_kwargs = {
            "name": name,
            "display_name": display_name,
            **kwargs,
        }
        return func
    return decorator


def datetime_parameter(name: str, display_name: str, include_time: bool = False, **kwargs):
    def decorator(func):
        func._ui_type = DateTimeParameter
        func._ui_kwargs = {
            "name": name,
            "display_name": display_name,
            "include_time": include_time,
            **kwargs
        }
        return func
    return decorator



import logging

from collections import namedtuple
from typing import Optional

from .grpc_stubs import tunnel_iface_pb2 as stubs, tunnel_iface_pb2_grpc as grpc_stubs
from ..grpc_interface import GRPCInterface
from ...utils import maybe_async
from ...cli.decorators import command as cli_command, annotate_arg

log = logging.getLogger(__name__)
Tunnel = namedtuple("Tunnel", ["address", "url"])


class TunnelInterface(GRPCInterface):
    stub = grpc_stubs.TunnelInterfaceStub

    def __init__(self, tunnel_uri: str = "127.0.0.1:50056", is_async: bool = None):
        super().__init__(tunnel_uri, is_async)

    def process_response(self, stub_call: str, response, *args, **kwargs):
        response = super().process_response(stub_call, response, *args, **kwargs)

        try:
            check_success = kwargs.pop("check_success")
        except KeyError:
            return response

        if check_success and not response.success:
            log.warning(f"Failed to execute {stub_call}. Process returned success failed. {response.message}")
            return None

        return response


    @cli_command()
    @maybe_async()
    def open_tunnel(
        self,
        address: str,
        protocol: str = "http",
        timeout: int = 15,
        username: str = None,
        password: str = None,
        domain: str = None,
    ) -> Optional[str]:
        """Open a tunnel with the given address and protocol.

        Returns the tunnel URL if opening the tunnel succeeded, None otherwise.
        """
        req = stubs.OpenTunnelRequest(tunnel=stubs.TunnelRequest(
            address=address,
            protocol=protocol,
            timeout=timeout,
            username=username,
            password=password,
            domain=domain,
        ))
        resp = self.make_request("OpenTunnel", req)
        return resp and resp.url

    async def open_tunnel_async(
        self,
        address: str,
        protocol: str = "http",
        timeout: int = 15,
        username: str = None,
        password: str = None,
        domain: str = None,
    ) -> Optional[str]:
        req = stubs.OpenTunnelRequest(tunnel=stubs.TunnelRequest(
            address=address,
            protocol=protocol,
            timeout=timeout,
            username=username,
            password=password,
            domain=domain,
        ))
        resp = await self.make_request_async("OpenTunnel", req)
        return resp and resp.url

    @cli_command()
    @maybe_async()
    def close_tunnel(self, address: str = None, url: str = None) -> bool:
        """Close tunnel by either address (ie. localhost:5000) or url (ie. https://random.ngrok.io)"""
        if not (address or url):
            raise ValueError("Address or URL required")

        req = stubs.CloseTunnelRequest(address=address, url=url)
        return self.make_request("CloseTunnel", req, check_success=True)

    async def close_tunnel_async(self, address: str = None, url: str = None) -> bool:
        if not (address or url):
            raise ValueError("Address or URL required")

        req = stubs.CloseTunnelRequest(address=address, url=url)
        return await self.make_request_async("CloseTunnel", req, check_success=True)

    @cli_command()
    @maybe_async()
    def close_all_tunnels(self) -> bool:
        """Close all open tunnels. Returns True if this succeeded, False otherwise."""
        return self.make_request("CloseAllTunnels", stubs.CloseTunnelRequest(), check_success=True)

    async def close_all_tunnels_async(self) -> bool:
        """Close all open tunnels. Returns True if this succeeded, False otherwise."""
        return await self.make_request_async("CloseAllTunnels", stubs.CloseTunnelRequest(), check_success=True)

    @cli_command()
    @maybe_async()
    def get_tunnel(self, address: str = None, url: str = None) -> Optional[Tunnel]:
        """Get tunnel by either address (ie. localhost:5000) or url (ie. https://random.ngrok.io)

        Returns a namedtuple with address and url attributes.
        """
        if not (address or url):
            raise ValueError("Address or URL required")

        resp = self.make_request("GetTunnel", stubs.GetTunnelRequest(address=address, url=url))
        return resp and Tunnel(resp.tunnel.address, resp.tunnel.url)

    async def get_tunnel_async(self, address: str = None, url: str = None) -> Optional[Tunnel]:
        if not (address or url):
            raise ValueError("Address or URL required")

        resp = await self.make_request_async("GetTunnel", stubs.GetTunnelRequest(address=address, url=url))
        return resp and Tunnel(resp.tunnel.address, resp.tunnel.url)

    @cli_command()
    @maybe_async()
    def get_all_tunnels(self) -> Optional[list[Tunnel]]:
        """Get all open tunnels. Returns a list of Tunnels with address and url attributes."""
        resp = self.make_request("ListTunnels", stubs.ListTunnelsRequest())
        return resp and [Tunnel(r.address, r.url) for r in resp.tunnels]

    async def get_all_tunnels_async(self) -> Optional[list[Tunnel]]:
        resp = await self.make_request_async("ListTunnels", stubs.ListTunnelsRequest())
        return resp and [Tunnel(r.address, r.url) for r in resp.tunnels]
    
    @cli_command()
    def test_comms(self, message: str = "Comms Check Message") -> Optional[str]:
        """Test connection by sending a basic echo response to tunnel interface container.

        Parameters
        ----------
        message : str
            Message to send to tunnel interface to have echo'd as a response

        Returns
        -------
        str
            The response from tunnel interface.
        """
        return self.make_request("TestComms", stubs.TestCommsRequest(message=message), response_field="response")




import enum
from datetime import datetime

from typing import Optional, Any

from .element import Element
from .submodule import Container


class Camera(Container):
    type = "uiCamera"

    def __init__(
        self,
        name,
        display_name: str = None,
        uri: str = None,
        output_type: str = None,
        mp4_output_length: int = None,
        wake_delay: int = 5,
        children: Optional[list[Element]] = None,
        **kwargs,
    ):
        super().__init__(name, display_name, children=children, **kwargs, is_available=None, help_str=None)
        # fixme: do we need to specify is_available and help_str to be None?

        self.uri = uri
        self.output_type = output_type
        self.mp4_output_length = mp4_output_length
        self.wake_delay = wake_delay

class CameraHistory(Camera):
    type = "uiCameraFeed"


doover_ui_camera = Camera
doover_ui_camera_history = CameraHistory


import inspect
import re
from typing import Any, Optional

from .element import Element


NAME_VALIDATOR = re.compile(r"^[a-zA-Z0-9_]+$")


class Container(Element):
    type = "uiContainer"

    def __init__(self, name, display_name=None, children: list[Element] = None, status_icon: str = None, auto_add_elements: bool = True, **kwargs):
        super().__init__(name, display_name, **kwargs)

        self._default_position = 101
        self._max_position = self._default_position

        # A list of doover_ui_elements
        self._children = dict()
        self.add_children(*children or [])

        self.status_icon = status_icon

        self._register_interactions()
        if auto_add_elements:
            self.add_children(*[e for name, e in inspect.getmembers(self, predicate=lambda e: isinstance(e, Element))])

    def _register_interactions(self):
        for name, func in inspect.getmembers(self, predicate=lambda f: inspect.ismethod(f) and hasattr(f, "_ui_type")):
            item = func._ui_type(**func._ui_kwargs)
            item.callback = func
            setattr(self, func.__name__, item)

    def to_dict(self):
        result = super().to_dict()

        if self.status_icon is not None:
            result['statusIcon'] = self.status_icon

        result["children"] = {name: c.to_dict() for name, c in self._children.items()}
        return result

    def get_diff(self, other: dict[str, Any], remove: bool = True, retain_fields: Optional[list] = []) -> Optional[dict[str, Any]]:
        res = super().get_diff(other, remove=remove, retain_fields=retain_fields) or {}
        # this will account for all the "normal" attributes, but not the children, since dicts aren't hashable
        # (ie. you can't do dict1 == dict2 to see if they're equal)
        other_children = other.get("children", {})
        this_children = {name: c for name, c in self._children.items()}

        children_diff = dict()
        if remove:
            children_diff.update({k: None for k in other_children if k not in this_children})  # to_remove

        for name, child in this_children.items():
            try:
                diff = child.get_diff(other_children[name], remove=remove, retain_fields=retain_fields)
                if diff is not None:
                    children_diff[name] = diff
            except KeyError:
                children_diff[name] = child.to_dict()

        if children_diff:
            res["children"] = children_diff

        if len(res) == 0:
            return None

        return res

    @property
    def children(self):
        return list(self._children.values())

    def set_children(self, children: list[Element]):
        self._children.clear()
        self._max_position = self._default_position
        self.add_children(*children)

    def add_children(self, *children: Element):

        # if not hasattr(self, "_default_position"):
        #     self._default_position = 101
        # if not hasattr(self, "_max_position"):
        #     self._max_position = self._default_position

        for c in children:
            if not isinstance(c, Element):
                continue

            name = c.name.strip()
            if not NAME_VALIDATOR.match(name):
                raise RuntimeError(f"Invalid name '{name}' for element '{c}'. Valid characters include letters, numbers, and underscores.")

            self._children[name] = c
            c.parent = self

            if not c.position:
                c.position = self._max_position
                self._max_position += 1

        return self

    def remove_children(self, *children: Element):
        for c in children:
            try:
                if c.name in self._children:
                    del self._children[c.name]
            except KeyError:
                pass
        
        ## for all self._children, call remove_children on them
        for c in self._children.values():
            if isinstance(c, Container):
                c.remove_children(*children)

    def clear_children(self):
        self._children.clear()

    def get_element(self, element_name: str) -> Optional[Element]:
        try:
            return self._children[element_name]
        except KeyError:
            pass

        for element in self._children.values():
            if isinstance(element, Container):
                elem = element.get_element(element_name)
                if elem is not None:
                    return elem


class Submodule(Container):
    type = "uiSubmodule"

    def __init__(
        self, name: str, display_name: str, children: list[Element] = None, status: str = None, is_collapsed: bool = False, **kwargs
    ):
        super().__init__(name, display_name, children, **kwargs)

        self.status = status or kwargs.pop("status_string", None)
        self.collapsed = is_collapsed or kwargs.pop("collapsed", False)

    def to_dict(self):
        result = super().to_dict()
        if self.status is not None:
            result['statusString'] = self.status
        # result['isCollapsed'] = self.collapsed()

        return result


doover_ui_container = Container
doover_ui_submodule = Submodule



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: grpc_stubs/tunnel_iface.proto
# Protobuf Python Version: 5.26.1
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x1dgrpc_stubs/tunnel_iface.proto\x12\x0ctunnel_iface\"q\n\x0eResponseHeader\x12\x0f\n\x07success\x18\x01 \x01(\x08\x12\x1a\n\rresponse_code\x18\x03 \x01(\x05H\x00\x88\x01\x01\x12\x14\n\x07message\x18\x04 \x01(\tH\x01\x88\x01\x01\x42\x10\n\x0e_response_codeB\n\n\x08_message\"#\n\x10TestCommsRequest\x12\x0f\n\x07message\x18\x01 \x01(\t\"\\\n\x11TestCommsResponse\x12\x35\n\x0fresponse_header\x18\x01 \x01(\x0b\x32\x1c.tunnel_iface.ResponseHeader\x12\x10\n\x08response\x18\x02 \x01(\t\"\x9e\x02\n\rTunnelRequest\x12\x14\n\x07\x61\x64\x64ress\x18\x01 \x01(\tH\x00\x88\x01\x01\x12\x15\n\x08protocol\x18\x02 \x01(\tH\x01\x88\x01\x01\x12\x14\n\x07timeout\x18\x03 \x01(\x05H\x02\x88\x01\x01\x12\x15\n\x08username\x18\x04 \x01(\tH\x03\x88\x01\x01\x12\x15\n\x08password\x18\x05 \x01(\tH\x04\x88\x01\x01\x12\x13\n\x06\x64omain\x18\x06 \x01(\tH\x05\x88\x01\x01\x12\x18\n\x0bremote_addr\x18\x07 \x01(\tH\x06\x88\x01\x01\x12\x13\n\x0b\x61llow_cidrs\x18\x08 \x03(\tB\n\n\x08_addressB\x0b\n\t_protocolB\n\n\x08_timeoutB\x0b\n\t_usernameB\x0b\n\t_passwordB\t\n\x07_domainB\x0e\n\x0c_remote_addr\"L\n\x0eTunnelResponse\x12\x14\n\x07\x61\x64\x64ress\x18\x01 \x01(\tH\x00\x88\x01\x01\x12\x10\n\x03url\x18\x02 \x01(\tH\x01\x88\x01\x01\x42\n\n\x08_addressB\x06\n\x04_url\"@\n\x11OpenTunnelRequest\x12+\n\x06tunnel\x18\x01 \x01(\x0b\x32\x1b.tunnel_iface.TunnelRequest\"P\n\x12\x43loseTunnelRequest\x12\x14\n\x07\x61\x64\x64ress\x18\x01 \x01(\tH\x00\x88\x01\x01\x12\x10\n\x03url\x18\x02 \x01(\tH\x01\x88\x01\x01\x42\n\n\x08_addressB\x06\n\x04_url\"\x18\n\x16\x43loseAllTunnelsRequest\"N\n\x10GetTunnelRequest\x12\x14\n\x07\x61\x64\x64ress\x18\x01 \x01(\tH\x00\x88\x01\x01\x12\x10\n\x03url\x18\x02 \x01(\tH\x01\x88\x01\x01\x42\n\n\x08_addressB\x06\n\x04_url\"\x14\n\x12ListTunnelsRequest\"v\n\x12OpenTunnelResponse\x12\x35\n\x0fresponse_header\x18\x01 \x01(\x0b\x32\x1c.tunnel_iface.ResponseHeader\x12\x0f\n\x07success\x18\x02 \x01(\x08\x12\x10\n\x03url\x18\x03 \x01(\tH\x00\x88\x01\x01\x42\x06\n\x04_url\"]\n\x13\x43loseTunnelResponse\x12\x35\n\x0fresponse_header\x18\x01 \x01(\x0b\x32\x1c.tunnel_iface.ResponseHeader\x12\x0f\n\x07success\x18\x02 \x01(\x08\"a\n\x17\x43loseAllTunnelsResponse\x12\x35\n\x0fresponse_header\x18\x01 \x01(\x0b\x32\x1c.tunnel_iface.ResponseHeader\x12\x0f\n\x07success\x18\x02 \x01(\x08\"x\n\x11GetTunnelResponse\x12\x35\n\x0fresponse_header\x18\x01 \x01(\x0b\x32\x1c.tunnel_iface.ResponseHeader\x12,\n\x06tunnel\x18\x02 \x01(\x0b\x32\x1c.tunnel_iface.TunnelResponse\"{\n\x13ListTunnelsResponse\x12\x35\n\x0fresponse_header\x18\x01 \x01(\x0b\x32\x1c.tunnel_iface.ResponseHeader\x12-\n\x07tunnels\x18\x02 \x03(\x0b\x32\x1c.tunnel_iface.TunnelResponse\"\x14\n\x12SyncTunnelsRequest\"]\n\x13SyncTunnelsResponse\x12\x35\n\x0fresponse_header\x18\x01 \x01(\x0b\x32\x1c.tunnel_iface.ResponseHeader\x12\x0f\n\x07success\x18\x02 \x01(\x08\x32\xe8\x04\n\x0fTunnelInterface\x12N\n\tTestComms\x12\x1e.tunnel_iface.TestCommsRequest\x1a\x1f.tunnel_iface.TestCommsResponse\"\x00\x12Q\n\nOpenTunnel\x12\x1f.tunnel_iface.OpenTunnelRequest\x1a .tunnel_iface.OpenTunnelResponse\"\x00\x12T\n\x0b\x43loseTunnel\x12 .tunnel_iface.CloseTunnelRequest\x1a!.tunnel_iface.CloseTunnelResponse\"\x00\x12`\n\x0f\x43loseAllTunnels\x12$.tunnel_iface.CloseAllTunnelsRequest\x1a%.tunnel_iface.CloseAllTunnelsResponse\"\x00\x12N\n\tGetTunnel\x12\x1e.tunnel_iface.GetTunnelRequest\x1a\x1f.tunnel_iface.GetTunnelResponse\"\x00\x12T\n\x0bSyncTunnels\x12 .tunnel_iface.SyncTunnelsRequest\x1a!.tunnel_iface.SyncTunnelsResponse\"\x00\x12T\n\x0bListTunnels\x12 .tunnel_iface.ListTunnelsRequest\x1a!.tunnel_iface.ListTunnelsResponse\"\x00\x62\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'grpc_stubs.tunnel_iface_pb2', _globals)
if not _descriptor._USE_C_DESCRIPTORS:
  DESCRIPTOR._loaded_options = None
  _globals['_RESPONSEHEADER']._serialized_start=47
  _globals['_RESPONSEHEADER']._serialized_end=160
  _globals['_TESTCOMMSREQUEST']._serialized_start=162
  _globals['_TESTCOMMSREQUEST']._serialized_end=197
  _globals['_TESTCOMMSRESPONSE']._serialized_start=199
  _globals['_TESTCOMMSRESPONSE']._serialized_end=291
  _globals['_TUNNELREQUEST']._serialized_start=294
  _globals['_TUNNELREQUEST']._serialized_end=580
  _globals['_TUNNELRESPONSE']._serialized_start=582
  _globals['_TUNNELRESPONSE']._serialized_end=658
  _globals['_OPENTUNNELREQUEST']._serialized_start=660
  _globals['_OPENTUNNELREQUEST']._serialized_end=724
  _globals['_CLOSETUNNELREQUEST']._serialized_start=726
  _globals['_CLOSETUNNELREQUEST']._serialized_end=806
  _globals['_CLOSEALLTUNNELSREQUEST']._serialized_start=808
  _globals['_CLOSEALLTUNNELSREQUEST']._serialized_end=832
  _globals['_GETTUNNELREQUEST']._serialized_start=834
  _globals['_GETTUNNELREQUEST']._serialized_end=912
  _globals['_LISTTUNNELSREQUEST']._serialized_start=914
  _globals['_LISTTUNNELSREQUEST']._serialized_end=934
  _globals['_OPENTUNNELRESPONSE']._serialized_start=936
  _globals['_OPENTUNNELRESPONSE']._serialized_end=1054
  _globals['_CLOSETUNNELRESPONSE']._serialized_start=1056
  _globals['_CLOSETUNNELRESPONSE']._serialized_end=1149
  _globals['_CLOSEALLTUNNELSRESPONSE']._serialized_start=1151
  _globals['_CLOSEALLTUNNELSRESPONSE']._serialized_end=1248
  _globals['_GETTUNNELRESPONSE']._serialized_start=1250
  _globals['_GETTUNNELRESPONSE']._serialized_end=1370
  _globals['_LISTTUNNELSRESPONSE']._serialized_start=1372
  _globals['_LISTTUNNELSRESPONSE']._serialized_end=1495
  _globals['_SYNCTUNNELSREQUEST']._serialized_start=1497
  _globals['_SYNCTUNNELSREQUEST']._serialized_end=1517
  _globals['_SYNCTUNNELSRESPONSE']._serialized_start=1519
  _globals['_SYNCTUNNELSRESPONSE']._serialized_end=1612
  _globals['_TUNNELINTERFACE']._serialized_start=1615
  _globals['_TUNNELINTERFACE']._serialized_end=2231
# @@protoc_insertion_point(module_scope)



import enum
from datetime import datetime

from typing import Optional, Any

from .misc import Colour


class Element:
    type = "uiElement"

    def __init__(
        self,
        name: Optional[str],
        display_name: Optional[str] = None,
        is_available: bool = None,  # not sure of type
        help_str: str = None,
        verbose_str: str = None,
        show_activity: bool = True,
        form: str = None,
        graphic: str = None,  # not sure of type
        layout: str = None,  # not sure of type
        component_url: str = None,  # not sure of type
        position: Optional[int] = None, # 100,
        conditions: Optional[dict] = None,
        **kwargs
    ):
        self.name = name
        self.display_name = display_name or kwargs.get("display_str", None)  # backwards compatibility
        self.is_available = is_available
        self.help_str = help_str
        self.verbose_str = verbose_str
        self.show_activity = show_activity
        self.form = form
        self.graphic = graphic
        self.layout = layout
        self.component_url = component_url
        self.position = position
        self.conditions = conditions

        self._retain_fields = []

    def to_dict(self):
        to_return = {
            "name": self.name,
            "type": self.type,
            "displayString": self.display_name,
            "isAvailable": self.is_available,
            "helpString": self.help_str,
            "verboseString": self.verbose_str,
            "showActivity": self.show_activity,
            "form": self.form,
            "graphic": self.graphic,
            "layout": self.layout,
            "componentUrl": self.component_url,
            "position": self.position,
            "conditions": self.conditions,
        }
        # filter out any null values
        return {k: v for k, v in to_return.items() if v is not None}

    def get_diff(self, other: dict[str, Any], remove: bool = True, retain_fields: Optional[list] = None) -> Optional[dict[str, Any]]:
        this = self.to_dict()

        retain_fields = retain_fields or []
        to_retain = list(set(retain_fields) | set(self._retain_fields))
        # if this == other:
        #     return None

        result = {k: v for k, v in this.items() if other.get(k) != v or k in to_retain}
        if remove:
            result.update(**{k: None for k in other if k not in this and k not in to_retain})  # to_remove
        if len(result) == 0:
            return None

        return result

    ## A stub for the method that will be called when the UI state is updated.
    # The element can choose to update its internal state based on the previous state and the new state.
    def recv_ui_state_update(self, state: dict[str, Any]) -> None:
        pass

class ConnectionType(enum.Enum):
    constant = "constant"
    periodic = "periodic"
    other = "other"


class ConnectionInfo(Element):
    """Connection Info

    Parameters
    ----------
    name: str
    connection_type: ConnectionType

    connection_period: int
        The expected time between connection events (seconds) - only applicable for "periodic"
    next_connection: int
        Expected time of next connection (seconds after shutdown)
    offline_after: int
        Show as offline if disconnected for more than x secs

    """
    type = "uiConnectionInfo"

    def __init__(
        self,
        name: str = "connectionInfo",
        connection_type: ConnectionType = ConnectionType.constant,
        connection_period: int = None,
        next_connection: int = None,
        offline_after: int = None,
        allowed_misses: int = None,
        **kwargs
    ):
        super().__init__(name, None, **kwargs)
        self.name = name
        self.connection_type = connection_type
        self.connection_period = connection_period
        self.next_connection = next_connection
        self.offline_after = offline_after
        self.allowed_misses = allowed_misses

        if self.connection_type is not ConnectionType.periodic and (
            self.connection_period is not None
            or self.next_connection is not None
            or self.allowed_misses is not None
        ):
            raise RuntimeError(
                "connection_type must be periodic to set connection_period, next_connection or offline_after"
            )

    def to_dict(self):
        result = {
            "name": self.name,
            "type": self.type,
            "connectionType": self.connection_type.value,
        }

        if self.connection_period is not None:
            result['connectionPeriod'] = self.connection_period
        if self.next_connection is not None:
            result['nextConnection'] = self.next_connection
        if self.offline_after is not None:
            result['offlineAfter'] = self.offline_after
        if self.allowed_misses is not None:
            result['allowedMisses'] = self.allowed_misses

        return result


class AlertStream(Element):
    type = "uiAlertStream"

    def __init__(self, name, display_name: str = None, **kwargs):
        super().__init__(name, display_name, is_available=None, help_str=None, **kwargs)

class Multiplot(Element):
    type = "uiMultiPlot"

    def __init__(
        self, name: str, display_name: str, series: list[str],
        series_colours: Optional[list[Colour]] = None, series_active: Optional[list[bool]] = None, 
        earliest_data_time: Optional[datetime] = None, title: Optional[str] = None,
        **kwargs
    ):
        super().__init__(name, display_name, **kwargs)

        self.series = series
        self.series_colours = series_colours
        self.series_active = series_active
        self.earliest_data_time = earliest_data_time
        self.title = title

    def to_dict(self):
        result = super().to_dict()
        result['series'] = self.series
        result['colours'] = self.series_colours
        result['activeSeries'] = self.series_active
        if self.title:
            result['title'] = self.title

        if self.earliest_data_time is not None:
            if isinstance(self.earliest_data_time, datetime):
                result['earliestDataDate'] = int(self.earliest_data_time.timestamp())
            else:
                result['earliestDataDate'] = self.earliest_data_time

        return result


class RemoteComponent(Element):
    type = "uiRemoteComponent"

    def __init__(self, name: str, display_name: str, component_url: str, **kwargs):
        super().__init__(name, display_name, component_url=component_url)
        self.kwargs = kwargs

    def to_dict(self):
        res = super().to_dict()
        res.update(self.kwargs)
        return res


doover_ui_element = Element
doover_ui_connection_info = ConnectionInfo
doover_ui_alert_stream = AlertStream
doover_ui_multiplot = Multiplot
doover_ui_remote_component = RemoteComponent




import threading
import time, logging, traceback, argparse, asyncio, signal
from typing import Optional, Any

from .camera import CameraManager, CameraInterface
from .device_agent import DeviceAgentInterface, device_agent_iface
from .modbus import ModbusInterface, modbus_iface
from .platform import PlatformInterface
from .tunnel import TunnelInterface

from .power_manager import PowerManager
from .location_manager import LocationManager

from ..ui import UIManager
from ..utils import maybe_async, CaseInsensitiveDict, call_maybe_async, get_is_async


log = logging.getLogger(__name__)

class app_manager_logging_formatter(logging.Formatter):

    grey = "\x1b[38;20m"
    yellow = "\x1b[33;20m"
    red = "\x1b[31;20m"
    bold_red = "\x1b[31;1m"
    reset = "\x1b[0m"
    format = "%(asctime)s - %(name)s - %(levelname)s - %(message)s (%(filename)s:%(lineno)d)"

    FORMATS = {
        logging.DEBUG: grey + format + reset,
        logging.INFO: grey + format + reset,
        logging.WARNING: yellow + format + reset,
        logging.ERROR: red + format + reset,
        logging.CRITICAL: bold_red + format + reset
    }

    def format(self, record):
        log_fmt = self.FORMATS.get(record.levelno)
        formatter = logging.Formatter(log_fmt)
        return formatter.format(record)


class app_base:

    def __init__(self, manager=None, name=None, is_async: bool = None):
        self.manager = manager

        self._is_async = get_is_async(is_async)
        self.ui_manager: Optional[UIManager] = None

        self._camera_manager: Optional[CameraManager] = None
        self._modbus_iface: Optional[modbus_iface] = None
        self._platform_iface: Optional[PlatformInterface] = None
        self._tunnel_iface: TunnelInterface = None

        self._power_manager: Optional[PowerManager] = None
        self._location_manager: Optional[LocationManager] = None

        if name is None:
            self.name = self.__class__.__name__
        else:
            self.name = name

        self.log_warnings_issued = []

    def set_manager(self, manager):
        self.manager = manager

    def get_app_name(self):
        return self.name

    def get_manager(self):
        return self.manager
    

    ## Agent Interface Functions (DDA)

    def get_agent_iface(self):
        return self.manager.get_agent_iface()
    
    def get_is_dda_available(self):
        return self.get_agent_iface().get_is_dda_available()
    
    def get_is_dda_online(self):
        return self.get_agent_iface().get_is_dda_online()
    
    def get_has_dda_been_online(self):
        return self.get_agent_iface().get_has_dda_been_online()

    def subscribe_to_channel(self, channel_name, callback=None):
        self.get_agent_iface().add_subscription(channel_name, callback)

    def publish_to_channel(self, channel_name, data):
        return self.get_agent_iface().publish_to_channel(channel_name, data)

    def get_channel_aggregate(self, channel_name):
        return self.get_agent_iface().get_channel_aggregate(channel_name)
    

    ## Config Manager Functions

    def get_config_manager(self) -> "deployment_config_manager":
        return self.manager.get_config_manager()
    
    def get_config(self, key_filter=None, wait=None, default=None):
        result = self.get_config_manager().get_config(key_filter, wait)
        if result is None:
            return default
        return result

    def get_config_async(self, key_filter=None, wait=None, default=None):
        result = self.get_config_manager().get_config_async(key_filter, wait)
        if result is None:
            return default
        return result

    ## UI Manager Functions

    def get_ui_manager(self):
        if self.ui_manager is None:
            self.ui_manager: UIManager = UIManager(agent_id=None, client=self.get_agent_iface(), is_async=self._is_async)
        return self.ui_manager
    
    def set_ui_elements(self, elements):
        return self.get_ui_manager().set_children(elements)

    def get_command(self, name):
        return self.get_ui_manager().get_command(name)
    
    def coerce_command(self, name, value):
        return self.get_ui_manager().coerce_command(name, value)
    
    def record_critical_value(self, name, value):
        return self.get_ui_manager().record_critical_value(name, value)

    def set_ui_status_icon(self, icon):
        return self.get_ui_manager().set_status_icon(icon)

    def start_ui_comms(self):
        ui_manager_obj = self.get_ui_manager()
        ui_manager_obj.start_comms()

    async def await_ui_comms_sync(self, timeout=10):
        logging.debug("Awaiting UI comms sync")
        result = await self.get_ui_manager().await_comms_sync(timeout=timeout)
        if result is False:
            logging.warning("UI comms sync timed out")
        else:
            logging.debug("UI comms sync complete")
        return result

    def set_ui(self, ui):
        self.get_ui_manager().set_children(ui)

    async def update_ui(self):
        await self.get_ui_manager().handle_comms_async()


    ## Platform Interface Functions

    @property
    def platform_iface(self):
        if self._platform_iface is None:
            self._platform_iface = self.get_manager().get_platform_iface()
        return self._platform_iface

    def get_platform_iface(self):
        return self.platform_iface

    def get_di(self, di):
        return self.platform_iface.get_di(di)

    def get_ai(self, ai):
        return self.platform_iface.get_ai(ai)

    def get_do(self, do):
        return self.platform_iface.get_do(do)
    
    def set_do(self, do, value):
        return self.platform_iface.set_do(do, value)
    
    def schedule_do(self, do, value, delay_secs):
        return self.platform_iface.schedule_do(do, value, delay_secs)
    
    def get_ao(self, ao):
        return self.platform_iface.get_ao(ao)

    def set_ao(self, ao, value):
        return self.platform_iface.set_ao(ao, value)
    
    def schedule_ao(self, ao, value, delay_secs):
        return self.platform_iface.schedule_ao(ao, value, delay_secs)
    

    ## Modbus Interface Functions

    @property
    def modbus_iface(self):
        if self._modbus_iface is None:
            self._modbus_iface = self.get_manager().get_modbus_iface()
        return self._modbus_iface

    def get_modbus_iface(self):
        return self.modbus_iface
        # return self.get_manager().get_modbus_iface()
    
    def read_modbus_registers(self, address, count, register_type, modbus_id=None, bus_id=None):
        return self.modbus_iface.read_registers(
            bus_id=bus_id, 
            modbus_id=modbus_id, 
            start_address=address,
            num_registers=count,
            register_type=register_type,
        )
    
    def write_modbus_registers(self, address, values, register_type, modbus_id=None, bus_id=None):
        return self.modbus_iface.write_registers(
            bus_id=bus_id, 
            modbus_id=modbus_id, 
            start_address=address,
            values=values,
            register_type=register_type,
        )
    
    def add_new_modbus_read_subscription(self, address, count, register_type, callback, poll_secs=None, modbus_id=None, bus_id=None):
        return self.modbus_iface.add_new_read_subscription(
            bus_id=bus_id, 
            modbus_id=modbus_id, 
            start_address=address,
            num_registers=count,
            register_type=register_type,
            poll_secs=poll_secs,
            callback=callback,
        )


    ## Camera Manager Functions

    @property
    def camera_manager(self):
        if not self._camera_manager:
            logging.info("Creating camera manager")
            self._camera_manager = CameraManager(
                config_manager=self.get_config_manager(),
                dda_iface=self.get_agent_iface(),
                ui_manager=self.ui_manager,
                plt_iface=self.platform_iface,
                app_manager=self.get_manager(),
            )
        return self._camera_manager

    def get_camera_manager(self):
        return self.camera_manager

    @property
    def tunnel_iface(self):
        if not self._tunnel_iface:
            self._tunnel_iface = self.get_manager().tunnel_iface
        return self._tunnel_iface

    ## Power Manager Functions

    @property
    def power_manager(self):
        if not self._power_manager:
            self._power_manager = PowerManager(app=self)
        return self._power_manager

    def register_request_shutdown_hook(self, hook):
        self.power_manager.register_request_shutdown_hook(hook)

    def register_pre_shutdown_hook(self, hook):
        self.power_manager.register_pre_shutdown_hook(hook)

    def register_shutdown_hook(self, hook):
        self.power_manager.register_shutdown_hook(hook)


    ## Location Manager Functions

    @property
    def location_manager(self):
        if not self._location_manager:
            self._location_manager = LocationManager(app=self)
        return self._location_manager
    

    ## App Functions

    async def _setup(self):
        log.info("Setting up internal app : " + self.get_app_name())
        self.start_ui_comms()
        await self.get_config_manager().await_config()
        await self.camera_manager.setup()
        await self.power_manager.setup()
        await self.location_manager.setup()
        await self.await_ui_comms_sync(timeout=15)

    async def _main_loop(self):
        log.debug("Running internal main_loop : " + self.get_app_name())
        await self.camera_manager.main_loop()
        await self.power_manager.main_loop()
        await self.location_manager.main_loop()
        await self.update_ui()

    def setup(self):
        return NotImplemented

    def main_loop(self):
        return NotImplemented

    def log(self, log_level="info", msg=None):

        ## Allows multiple variants of arguments to be accepted as valid
        if msg is None and log_level is not None:
            msg = log_level
            log_level = None

        if log_level not in ["debug", "info", "warning", "error", "critical"]:
            if log_level not in self.log_warnings_issued:
                logging.warning("Invalid log level '" + str(log_level) + "' , defaulting to 'info'")
                self.log_warnings_issued.append(log_level)
            log_level = "info"
        log_function = getattr(logging, log_level)
        log_function("[" + self.get_app_name() + "] " + str(msg))

    def close_app(self, with_delay=None):
        self.manager.close(with_delay=with_delay)

class deployment_config_manager:
    def __init__(self, dda_iface=None, auto_start=True, update_callback=None):
        self.dda_iface = dda_iface
        
        self.deployment_channel_name = 'deployment_config'

        self.is_subscribed = False
        self.has_recv_message = False
        self.last_deployment_config = {}
        self.update_callback = update_callback

        self.is_ready_event = asyncio.Event()

        if auto_start and self.dda_iface is not None:
            self.setup()


    def set_dda_iface(self, dda_iface):
        self.dda_iface = dda_iface

    def setup(self):
        if self.is_subscribed:
            return ## already setup

        self.is_subscribed = True
        self.dda_iface.add_subscription(self.deployment_channel_name, self.recv_updates)

    async def await_config(self, wait_period: int = 30):
        logging.info("Awaiting deployment config...")

        self.setup()
        try:
            await asyncio.wait_for(self.is_ready_event.wait(), wait_period)
            logging.info("Received deployment config")
        except asyncio.TimeoutError:
            logging.warning("Failed to receive deployment config due to timeout.")

        return self.last_deployment_config

    async def recv_updates(self, channel_name, last_aggregate):
        if channel_name is not self.deployment_channel_name:
            return ## Not the correct channel - something weird here
        
        if 'deployment_config' not in last_aggregate:
            logging.info("No deployment_config field in last deployment_config channel aggregate")
            return

        self.last_deployment_config = CaseInsensitiveDict.from_dict(last_aggregate["deployment_config"])
        logging.debug(f"Received deployment config: {self.last_deployment_config}")

        self.has_recv_message = True
        self.is_ready_event.set()

        if self.update_callback is not None:
            await call_maybe_async(self.update_callback, self.last_deployment_config)


    def _do_get_config(self, key_filter: str | list[str]) -> Optional[Any]:
        # When this goes up for review feel free to disagree with me,
        # but I don't see a reason for case sensitivity in the config lookup
        # I'm pretty sure the site turns everything to UPPER anyway
        config = self.last_deployment_config
        if config is None or len(config.keys()) == 0:
            return None

        if isinstance(key_filter, str):
            # fast route
            try:
                return config[key_filter.lower()]
            except KeyError:
                return None

        key_filter = [k.lower() for k in key_filter if isinstance(k, str)]
        if not key_filter:
            return None

        # fixme: not sure what to do here, the old code would return the
        #  last matching key, and if any didn't exist return None.

        # in the interim I'm just going to return the first matching instance.
        for key in key_filter:
            try:
                return config[key]
            except KeyError:
                continue

    @maybe_async()
    def get_config(self, key_filter=None, wait=True, wait_period=10, case_sensitive=False):
        if wait and not self.has_recv_message:
            wait_start = time.time()
            while not self.has_recv_message and (time.time() - wait_start < wait_period):
                time.sleep(0.25)

        return self._do_get_config(key_filter)

    async def get_config_async(self, key_filter=None, wait=True, wait_period=10, case_sensitive=False):
        if wait and not self.has_recv_message:
            await self.await_config(wait_period)

        return self._do_get_config(key_filter)


class app_manager:

    def __init__(
        self,
        loop_obj=None,
        device_agent=None,
        platform_iface=None,
        modbus_iface=None,
        camera_iface=None,
        tunnel_iface: TunnelInterface = None
    ):
        self.loop_obj = loop_obj
        self.device_agent = device_agent
        self.platform_iface = platform_iface
        self.modbus_iface = modbus_iface
        self.camera_iface = camera_iface
        self.tunnel_iface = tunnel_iface

        self.restart_all_on_error = True
        self.error_wait_period = 10
        self.should_stop = False
        self.setup_functions = []
        self.loop_functions = []

        if self.loop_obj is None:
            self.create_loop()
        if self.device_agent is None:
            self.device_agent = DeviceAgentInterface()

        self.config_manager = deployment_config_manager(dda_iface=self.device_agent)

        if self.platform_iface is None:
            self.platform_iface = platform_iface()


    def register_loop(self, callable):
        self.loop_functions.append(callable)

    def register_setup_function(self, callable):
        self.setup_functions.append(callable)


    async def main_loop(self):

        ## ensure config is recieved before starting
        await self.config_manager.await_config()

        ## trigger modbus iface setup
        self.modbus_iface.set_config_manager(self.config_manager)
        await self.modbus_iface.setup_from_config_manager()

        while not self.should_stop:
            restart_loop = False

            for setup in self.setup_functions:
                try:
                    ## Don't run setup in executor, to give the user the ability to setup async stuff in the setup function
                    await call_maybe_async(setup, in_executor=False)
                except Exception as e:
                    logging.error("Error in setup function: " + str(e))
                    # logging.error("Setup: " + str(setup))
                    logging.error(traceback.format_exc())
                    if self.restart_all_on_error:
                        logging.warning("\n\n\nWaiting " + str(self.error_wait_period) + " seconds before closing app\n\n")
                    else:
                        logging.warning("\n\n\nWaiting " + str(self.error_wait_period) + " seconds before restarting app\n\n")
                    await asyncio.sleep(self.error_wait_period)
                    
                    restart_loop = True
                    break

            ## allow for other async tasks to run between setup and loop
            await asyncio.sleep(0.2)

            while not self.should_stop and not restart_loop:
                for loop in self.loop_functions:
                    try:
                        await call_maybe_async(loop)
                    except Exception as e:
                        logging.error("Error in loop function: " + str(e))
                        # logging.error("Loop: " + str(loop))
                        logging.error(traceback.format_exc())
                        if self.restart_all_on_error:
                            logging.warning("\n\n\nWaiting " + str(self.error_wait_period) + " seconds before closing app\n\n")
                        else:
                            logging.warning("\n\n\nWaiting " + str(self.error_wait_period) + " seconds before restarting app\n\n")
                        await asyncio.sleep(self.error_wait_period)

                        restart_loop = True
                        break

                ## Allow for other async tasks to run
                await asyncio.sleep(0.2)

            if self.restart_all_on_error:
                self.close()
                return

    def start_task(self, function):
        logging.warning("app_manager.start_task is deprecated. Use an async task instead!")
        thread = threading.Thread(target=asyncio.run, args=(function(), ))
        thread.start()
        return thread

    def create_loop(self):
        if self.loop_obj is None:
            self.loop_obj = asyncio.get_event_loop()
            
            self.loop_obj.add_signal_handler(signal.SIGINT, self.close)
            self.loop_obj.add_signal_handler(signal.SIGTERM, self.close)

            asyncio.set_event_loop(self.loop_obj)

    def get_loop(self):
        return self.loop_obj
    
    def get_agent_iface(self):
        return self.device_agent

    def get_config_manager(self):
        return self.config_manager

    @maybe_async()
    def get_config(self, key_filter: str, wait: bool = True):
        return self.config_manager.get_config(key_filter=key_filter, wait=wait)

    async def get_config_async(self, key_filter: str, wait: bool = True):
        return await self.config_manager.get_config_async(key_filter=key_filter, wait=wait)

    def get_platform_iface(self):
        return self.platform_iface
    
    def get_modbus_iface(self):
        return self.modbus_iface
    
    def get_camera_iface(self):
        return self.camera_iface

    async def run(self):
        try:
            await self.main_loop()
        except asyncio.CancelledError:
            logging.info("Main loop cancelled")
        except Exception as e:
            logging.error("Error in main loop: " + str(e))
            logging.error(traceback.format_exc())

    def close(self, with_delay=None):
        if not self.should_stop:
            logging.info("\n########################################\n\nClosing app manager...\n\n########################################\n")
            self.should_stop = True
            if self.device_agent is not None:
                self.device_agent.close()

            if self.platform_iface is not None:
                self.platform_iface.close()

            def stop_loop():
                loop_obj = asyncio.get_event_loop()
                tasks = asyncio.all_tasks(loop_obj)
                for task in tasks:
                    task.cancel()
                # loop_obj.stop()

            self.loop_obj.call_soon_threadsafe(stop_loop)

            if with_delay is not None:
                time.sleep(with_delay)


async def main(app, dda_iface=None, plt_iface=None, mb_iface=None, cam_iface=None, tunnel_iface: TunnelInterface = None, debug=False):
    parser = argparse.ArgumentParser(description='Doover Docker App Manager')

    parser.add_argument('--remote-dev', type=str, default=None, help='Remote device URI')
    parser.add_argument('--dda-uri', type=str, default="localhost:50051", help='Doover Device Agent URI')
    parser.add_argument('--plt-uri', type=str, default="localhost:50053", help='Platform Interface URI')
    parser.add_argument('--modbus-uri', type=str, default="localhost:50054", help='Modbus Interface URI')
    parser.add_argument('--cam-uri', type=str, default="localhost:50055", help='Camera Interface URI')
    parser.add_argument('--tunnel-uri', type=str, default="localhost:50056", help='Tunnel Interface URI')
    parser.add_argument('--dds-sys-sock', type=str, default="/var/lib/dds/dds_sys.sock", help='DDS System Socket File Path')
    parser.add_argument('--debug', action='store_true', default=False, help='Debug Mode')

    args = parser.parse_args()

    if args.debug or debug:
        logging.basicConfig(level=logging.DEBUG)
    else:
        logging.basicConfig(level=logging.INFO)

    handler = logging.StreamHandler()
    handler.setFormatter(app_manager_logging_formatter())
    logging.getLogger().handlers.clear()
    logging.getLogger().addHandler(handler)

    dda_uri = args.dda_uri
    plt_uri = args.plt_uri
    modbus_uri = args.modbus_uri
    cam_uri = args.cam_uri
    tunnel_uri = args.tunnel_uri

    user_is_async = asyncio.iscoroutinefunction(app.setup) or asyncio.iscoroutinefunction(app.main_loop)
    app._is_async = get_is_async(user_is_async)

    if args.remote_dev is not None:
        dda_uri = args.dda_uri.replace("localhost", args.remote_dev)
        plt_uri = args.plt_uri.replace("localhost", args.remote_dev)
        modbus_uri = args.modbus_uri.replace("localhost", args.remote_dev)
        cam_uri = args.cam_uri.replace("localhost", args.remote_dev)
        tunnel_uri = args.tunnel_uri.replace("localhost", args.remote_dev)

    if dda_iface is None:
        dda_iface = device_agent_iface(dda_uri, user_is_async)

    if plt_iface is None:
        plt_iface = PlatformInterface(plt_uri, user_is_async)

    if mb_iface is None:
        mb_iface = ModbusInterface(modbus_uri, user_is_async)

    if cam_iface is None:
        cam_iface = CameraInterface(cam_uri, user_is_async)

    if tunnel_iface is None:
        tunnel_iface = TunnelInterface(tunnel_uri, user_is_async)

    app_manager_obj = app_manager(
        device_agent=dda_iface,
        platform_iface=plt_iface,
        modbus_iface=mb_iface,
        camera_iface=cam_iface,
        tunnel_iface=tunnel_iface,
    )
    app.set_manager(app_manager_obj)

    # register internal varients of setup and main_loop to stop possibility of user overriding them.
    # if they try hard enough they can override them but the average Joe shouldn't be...

    # this also allows us to raise an error if they haven't implemented either.
    app_manager_obj.register_setup_function(app._setup)
    app_manager_obj.register_loop(app._main_loop)

    app_manager_obj.register_setup_function(app.setup)
    app_manager_obj.register_loop(app.main_loop)

    await app_manager_obj.run()


def run_app(app, *args, **kwargs):
    asyncio.run(main(app, *args, **kwargs))


if __name__ == "__main__":

    run_app(app_base())


#!/usr/bin/env python3

import time, logging


## A simple 1D Kalman filter implementation
## This filter is designed to be used with a single sensor reading (e.g., a voltage reading, a temperature reading, etc.)
## It includes automatic time step calculation based on the time since the last update, however this can be overridden by providing a dt parameter.
## Has some basic outlier detection and handling

## If you want to adjust the overall sensitivity of the filter, you can adjust the process_variance parameter.
## Default is 0.5, which is a good starting point for most applications. 1 will make the filter more sensitive to changes, 0.1 will make it less sensitive.

## If you want to adjust the sensitivity to outliers, you can adjust the outlier_threshold which defaults to 5.
## Increase this value to make the filter less sensitive to outliers.

## If you are not familiar with Kalman filters, here is a good resource:
## https://www.bzarg.com/p/how-a-kalman-filter-works-in-pictures/

class KalmanFilter1D:

    def __init__(self, initial_estimate=None, initial_error_estimate=None, process_variance=None, outlier_protection=None, outlier_threshold=None, outlier_variance_multiplier=None):

        self.debug = False

        self.default_process_variance = 0.5  # Default process variance if not provided
        self.default_initial_estimate_ratio = 25 # The ratio of the initial estimate variance to the first measurement variance
        self.default_measurement_variance = 0.5  # Default measurement variance if not provided

        self.outlier_protection = outlier_protection if outlier_protection is not None else True  # Whether to use outlier protection
        self.default_outlier_threshold = 5  # The threshold for an outlier (in multiples of the error estimate)
        self.default_outlier_variance_multiplier = 25  # Multiplier for the measurement variance for an outlier

        self.estimate = initial_estimate  # Initial estimate (x_hat)
        self.error_estimate = initial_error_estimate  # Initial error estimate (P)
        self.process_variance = process_variance or self.default_process_variance  # Process variance (Q)

        self.outlier_threshold = outlier_threshold or self.default_outlier_threshold  # Threshold for an outlier
        self.outlier_variance_multiplier = outlier_variance_multiplier or self.default_outlier_variance_multiplier

        self.kalman_gain = 0  # Kalman gain (K)
        self.last_timestamp = None  # To store the last time the filter was updated
    
    def set_estimate(self, estimate):
        self.estimate = estimate

    def set_error_estimate(self, error_estimate):
        self.error_estimate = error_estimate

    def set_process_variance(self, process_variance):
        self.process_variance = process_variance

    def set_outlier_protection(self, outlier_protection):
        self.outlier_protection = outlier_protection

    def set_outlier_threshold(self, outlier_threshold):
        self.outlier_threshold = outlier_threshold

    def set_outlier_variance_multiplier(self, outlier_variance_multiplier):
        self.outlier_variance_multiplier = outlier_variance_multiplier

    def update(
            self,
            measurement,
            measurement_variance=None,
            dt=None,
            outlier_protection=None,
            process_variance=None,
        ):
        
        if process_variance is not None:
            if self.debug: logging.debug(f"Setting process variance to {process_variance}")
            self.process_variance = process_variance

        ## If the measurement is None, return the last estimate
        if measurement is None:
            return self.estimate
        
        ## If the measurement variance is not provided, set it to the default
        measurement_variance = measurement_variance or self.default_measurement_variance

        ## If not initialized, set the initial estimate to the first measurement
        if self.error_estimate is None:
            if self.debug: logging.debug("Initializing Kalman filter")
            self.error_estimate = measurement_variance * self.default_initial_estimate_ratio
        if self.estimate is None:
            self.estimate = measurement
            self.last_timestamp = time.time()
            return self.estimate

        current_time = time.time()
        # If dt is not provided, calculate it based on the time since the last update
        if dt is None:
            if self.last_timestamp is None:
                dt = 1  # Default to 1 second if this is the first call
            else:
                dt = current_time - self.last_timestamp
        self.last_timestamp = current_time

        # If the measurement is an outlier, increase the measurement variance
        outlier_protection = outlier_protection if outlier_protection is not None else self.outlier_protection
        outlier_message = None
        if outlier_protection and abs(measurement - self.estimate) > self.outlier_threshold * self.error_estimate:
            # logging.debug(f"Outlier detected: {measurement} (threshold: {self.outlier_threshold * self.error_estimate})")
            outlier_message = f"Outlier detected: {measurement} (threshold: {self.outlier_threshold * self.error_estimate})"
            measurement_variance *= self.outlier_variance_multiplier

        # Adjust process variance with dt
        adjusted_process_variance = self.process_variance * dt
        
        # Also adjust the process variance based on the magnitude of the measurement. This makes the filter more performant over datasets of wider magnitude.
        # adjusted_process_variance = self.process_variance * (1 + (abs(measurement) / 10)) ## This is a good starting point for most applications

        # Prediction step
        self.error_estimate += adjusted_process_variance

        # Update step
        if (self.error_estimate + measurement_variance) == 0:
            ## If the error estimate + measurement variance is 0, avoid division by zero
            self.kalman_gain = self.error_estimate / 0.0001
        else:
            self.kalman_gain = self.error_estimate / (self.error_estimate + measurement_variance)
        self.estimate += self.kalman_gain * (measurement - self.estimate)
        self.error_estimate = (1 - self.kalman_gain) * self.error_estimate

        debug_output = f"Measurement: {measurement}, Estimate: {self.estimate}, Error estimate: {self.error_estimate}"
        if outlier_message:
            debug_output += f" | {outlier_message}"
        if self.debug: logging.debug(debug_output)

        return self.estimate
    

## A decorator to apply a Kalman filter to the return value of a function
## The function should return a single value (e.g., a sensor reading)
## See below for an example of how to use this decorator
def apply_kalman_filter(
        initial_estimate=None,
        initial_error_estimate=None,
        process_variance=None,
        outlier_protection=None,
        outlier_threshold=None,
        outlier_variance_multiplier=None,    
    ):
    
    def decorator(func):

        ## Generate an id for the Kalman filter instance
        kf_id = id(func)

        def wrapper(self, *args, **kwargs):
            
            ## Instantiate the Kalman filter if it doesn't exist
            ## This allows multiple instances of the same function to have separate Kalman filters
            if not hasattr(self, '_kf_instances'):
                self._kf_instances = {}
            if kf_id not in self._kf_instances:
                # Create an instance of the Kalman filter with the provided parameters
                self._kf_instances[kf_id] = KalmanFilter1D(
                    initial_estimate=initial_estimate,
                    initial_error_estimate=initial_error_estimate,
                    process_variance=process_variance,
                    outlier_protection=outlier_protection,
                    outlier_threshold=outlier_threshold,
                    outlier_variance_multiplier=outlier_variance_multiplier,
                )
                
            _kalman_filter = self._kf_instances[kf_id]

            wrapper._kalman_filter = _kalman_filter

            # Call the original function and get its return value
            result = func(self, *args, **kwargs)
            
            # Extract measurement_variance and optional dt from kwargs or set defaults
            measurement_variance = kwargs.pop('kf_measurement_variance', None)
            dt = kwargs.pop('kf_dt', None)  # Allow dt to be optionally overridden
            measurement_outlier_protection = kwargs.pop('kf_outlier_protection', None)  # Allow outlier_protection to be optionally overridden
            kf_process_variance = kwargs.pop('kf_process_variance', None)  # Allow process_variance to be optionally overridden

            # Apply the Kalman filter to the return value
            filtered_result = _kalman_filter.update(result, measurement_variance, dt, measurement_outlier_protection, kf_process_variance)
            return filtered_result
        return wrapper
    
    return decorator


def apply_async_kalman_filter(
        initial_estimate=None,
        initial_error_estimate=None,
        process_variance=None,
        outlier_protection=None,
        outlier_threshold=None,
        outlier_variance_multiplier=None,    
    ):
    
    def decorator(func):

        ## Generate an id for the Kalman filter instance
        kf_id = id(func)

        async def wrapper(self, *args, **kwargs):
            
            ## Instantiate the Kalman filter if it doesn't exist
            ## This allows multiple instances of the same function to have separate Kalman filters
            if not hasattr(self, '_kf_instances'):
                self._kf_instances = {}
            if kf_id not in self._kf_instances:
                # Create an instance of the Kalman filter with the provided parameters
                self._kf_instances[kf_id] = KalmanFilter1D(
                    initial_estimate=initial_estimate,
                    initial_error_estimate=initial_error_estimate,
                    process_variance=process_variance,
                    outlier_protection=outlier_protection,
                    outlier_threshold=outlier_threshold,
                    outlier_variance_multiplier=outlier_variance_multiplier,
                )
                
            _kalman_filter = self._kf_instances[kf_id]

            wrapper._kalman_filter = _kalman_filter

            # Call the original function and get its return value
            result = await func(self, *args, **kwargs)
            
            # Extract measurement_variance and optional dt from kwargs or set defaults
            measurement_variance = kwargs.pop('kf_measurement_variance', None)
            dt = kwargs.pop('kf_dt', None)  # Allow dt to be optionally overridden
            measurement_outlier_protection = kwargs.pop('kf_outlier_protection', None)  # Allow outlier_protection to be optionally overridden
            kf_process_variance = kwargs.pop('kf_process_variance', None)  # Allow process_variance to be optionally overridden

            # Apply the Kalman filter to the return value
            filtered_result = _kalman_filter.update(result, measurement_variance, dt, measurement_outlier_protection, kf_process_variance)
            return filtered_result
        return wrapper
    
    return decorator



## Demo of using the apply_kalman_filter decorator
if __name__ == "__main__":

    class TestClass:

        def __init__(self):

            ## A list of return values simulating a battery voltage with an intermittent drop
            self.values = [
                12.0, 11.5, 11.0, 10.0, 10.5, 11.0, 11.5, 5, 12.0, 12.5, 13.0, 13.5, 14.0, 14.5, 15.0, 
                0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
                12, 12, 12, 12, 12, 12, 12, 12, 12, 12, 12, 12, 12, 12, 12, 12, 12, 12, 12, 12, 12, 12, 12, 
                None, None, None, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8
            ]

            ## An alternate list the same as the batteries but negative
            # self.values = [-v for v in self.values if v is not None]

            ## Another alternate list, but of a much smaller magnitude
            # self.values = [v/100 for v in self.values if v is not None]

            ## another alternate list, but with a much larger magnitude
            self.values_2 = [v*100 for v in self.values if v is not None]

            # ## A list of return values simulating a gas ppm sensor with an intermittent spike
            # self.values = [
            #     5000, 4900, 4800, 4700, 4850, 5000, 5150, 5300, 5250, 5200, 5150, 5100, 5050, 5000, 4950, 4900,
            #     20000, 19500, 19000, 500, 600, 550, 500, 500, 500, 500, 500, 500, 500, 500, 500, 500, 500,
            #     5000, 4900, 4800, 4700, 4850, 5000, 5150, 5300, 5250, 5200, 5150, 5100, 5050, 5000, 4950, 4900,
            #     0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
            # ]

            self.counter = 0

        ## define a generator function that returns the next sensor value
        def _next(self):
            # increment the counter
            self.counter += 1

        def is_running(self):
            return self.counter < len(self.values)

        def _raw_sensor_value(self):
            if self.counter < len(self.values):
                return self.values[self.counter]
            return None
        
        def _raw_sensor_value_2(self):
            if self.counter < len(self.values):
                return self.values_2[self.counter]
            return None

        ## Can optionally provide initial_estimate, initial_error_estimate, and process_variance, among others here also
        @apply_kalman_filter()
        def sensor_value(self, kf_process_variance=None):
            return self._raw_sensor_value()
        
        @apply_kalman_filter()
        def sensor_value_2(self):
            return self._raw_sensor_value_2()


    logging.basicConfig(level=logging.DEBUG)

    test = TestClass()
    # logging.info(f"Raw values: {test.values}")

    while test.is_running():

        filtered_value = test.sensor_value(kf_process_variance=0.001)
        # filtered_value_2 = test.sensor_value_2()
        # Can also provide measurement_variance and dt as keyword arguments. e.g.
        # filtered_value = test.sensor_value(kf_measurement_variance=0.5, kf_dt=1)

        # logging.info(f"Raw value: {test._raw_sensor_value()}, Filtered value: {filtered_value}")
        test._next()
        time.sleep(1)


import asyncio
import json
import logging

from collections.abc import MutableMapping

from functools import wraps


log = logging.getLogger(__name__)

## A function to map a reading to a value in a range
def map_reading(in_val, output_values, raw_readings=[4,20], ignore_below=3):

    if in_val < ignore_below:
        return None

    ## Choose the value set to map between
    lower_val_ind = 0
    found = False
    for i in range(0, len(raw_readings)):
        if in_val <= raw_readings[i]:
            lower_val_ind = i-1
            found = True
            break

    if not found:
        lower_val_ind = len(raw_readings)-2

    # Figure out how 'wide' each range is
    inSpan = raw_readings[lower_val_ind + 1] - raw_readings[lower_val_ind]
    outSpan = output_values[lower_val_ind + 1] - output_values[lower_val_ind]

    # Convert the left range into a 0-1 range (float)
    valueScaled = float(in_val - raw_readings[lower_val_ind]) / float(inSpan)

    # Convert the 0-1 range into a value in the right range.
    return output_values[lower_val_ind] + (valueScaled * outSpan)


def find_object_with_key(obj, key_to_find):
    stack = [obj]

    while stack:
        current = stack.pop()

        if isinstance(current, dict):
            if key_to_find in current:
                return current[key_to_find]

            for key in current:
                stack.append(current[key])

    return None


def find_path_to_key(obj, key_to_find):
    stack = [{'current': obj, 'path': ''}]

    while stack:
        current_entry = stack.pop()
        current = current_entry['current']
        path = current_entry['path']

        if isinstance(current, dict):
            if key_to_find in current:
                return f"{path}.{key_to_find}" if path else key_to_find

            for key in current:
                new_path = f"{path}.{key}" if path else key
                stack.append({'current': current[key], 'path': new_path})

    return None


def get_is_async(is_async: bool = None):
    if is_async is not None:
        return is_async

    try:
        asyncio.get_running_loop()
        return True
    except RuntimeError:
        return False


def maybe_async():
    def wrapper(func):
        @wraps(func)
        def inner(*args, **kwargs):
            # args[0] is self
            is_async_context = getattr(args[0], "_is_async", False)
            # allow them to specify if they want to run the sync version of the function (default very much not)
            force_sync = kwargs.pop("run_sync", False)

            if is_async_context is True and force_sync is False:
                # we're in an async context, check if we have an async variety of the function to run...
                try:
                    alternative = getattr(args[0], f"{func.__name__}_async")
                    return alternative(*args[1:], **kwargs)
                except AttributeError:
                    # we don't have a corresponding async method, just use the sync one.
                    pass
            return func(*args, **kwargs)
        return inner
    return wrapper


def wrap_try_except(func, *args, **kwargs):
    try:
        return func(*args, **kwargs)
    except Exception as e:
        log.exception(f"Error in {func.__name__}: {e}", exc_info=e)

async def wrap_try_except_async(func, *args, **kwargs):
    try:
        return await func(*args, **kwargs)
    except Exception as e:
        log.exception(f"Error in {func.__name__}: {e}", exc_info=e)

async def call_maybe_async(func, *args, as_task: bool = False, in_executor: bool = True, **kwargs):
    # print(f"call_maybe_async: func={func}, as_task={as_task}, in_executor={in_executor is True and not asyncio.iscoroutinefunction(func)}")
    if asyncio.iscoroutinefunction(func):
        # coro = wrap_try_except_async(func, *args, **kwargs)
        coro = func(*args, **kwargs)
        if as_task:
            # assign it to a variable for weak ref
            task = asyncio.create_task(coro)
            return task
        else:
            return await coro
    elif in_executor:
        loop = asyncio.get_running_loop()
        # run_in_executor doesn't support kwargs
        if kwargs:
            log.warning("kwargs are not supported when calling via executor")

        # this is a little bit of a hack, but essentially we're creating an async function that
        # is called with await func to allow for both running as a task and in an executor.
        # future = loop.run_in_executor(None, wrap_try_except, func, *args)
        future = loop.run_in_executor(None, func, *args, **kwargs)
        if as_task:
            return future

        return await future
    else:
        # return wrap_try_except(func, catch_exception, *args, **kwargs)
        return func(*args, **kwargs)


def on_change(callback, name=None):
    """
    A decorator that triggers a callback when the output of the decorated function changes.
    
    The callback is called with four arguments:
      - new_result: The new output of the function.
      - old_result: The previous output (or None if this is the first call).
      - is_first: A boolean indicating if this is the first time the function has returned a value.
      - change_detector_name: The optional name identifying this change detector.
    
    :param callback: The callback function to trigger, or a string indicating the name of an instance attribute.
    :param name: An optional name for this change_detector_name.

    Example:
    ```
    class MyClass:
        def __init__(self):
            self.last = None

        def my_callback(self, new_result, old_result, is_first, change_detector_name):
            if is_first:
                print(f"{change_detector_name} has returned a value for the first time: {new_result}")
            else:
                print(f"{change_detector_name} has changed from {old_result} to {new_result}")

        @on_change("my_callback", name="my_function")
        def my_function(self):
            import random
            return random.randint(0, 100)
    ``` 
    """
    def decorator(func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            detector_name = name or func.__name__

            # Check if we're dealing with an instance method.
            if args and hasattr(args[0], '__dict__'):
                instance = args[0]
                state_attr = f"__on_change_state_{detector_name}"
                # Retrieve state from the instance, or initialize it if not present.
                state = getattr(instance, state_attr, {"last_result": None, "has_value": False})
            else:
                # For non-instance functions, fallback to closure state.
                # (You could also raise an error if you only expect instance methods.)
                state = wrapper.__dict__.setdefault('state', {"last_result": None, "has_value": False})
            
            result = func(*args, **kwargs)
            is_first = not state["has_value"]
            
            # Trigger the callback if it's the first call or if the result has changed.
            if is_first or result != state["last_result"]:
                # Determine which callback to use.
                if isinstance(callback, str):
                    if not args:
                        raise ValueError("Expected a 'self' argument when using a string callback.")
                    cb = getattr(args[0], callback, None)
                    if cb is None or not callable(cb):
                        raise ValueError(f"Attribute '{callback}' is not a callable of the instance.")
                else:
                    cb = callback

                cb(result, state["last_result"] if state["has_value"] else None, is_first, detector_name)
                state["last_result"] = result
                state["has_value"] = True
                
                # Save the updated state back to the instance (or closure, as applicable).
                if args and hasattr(args[0], '__dict__'):
                    setattr(instance, state_attr, state)
                else:
                    wrapper.__dict__['state'] = state

            return result
        return wrapper
    return decorator

class CaseInsensitiveDict(MutableMapping):
    def __init__(self, data=None, **kwargs):
        self._store = dict()
        if data is None:
            data = {}
        self.update(data, **kwargs)

    def copy(self):
        return CaseInsensitiveDict(self._store)

    def to_dict(self):
        return {k: v.to_dict() if isinstance(v, CaseInsensitiveDict) else v for k, v in self.items()}

    @classmethod
    def from_dict(cls, data):
        t = {k: CaseInsensitiveDict.from_dict(v) if isinstance(v, dict) else v for k, v in data.items()}
        return cls(t)

    def __len__(self):
        return len(self._store)

    def __iter__(self):
        return iter(self._store)

    def __setitem__(self, key, value):
        self._store[key.lower()] = value

    def __getitem__(self, key):
        return self._store[key.lower()]

    def __delitem__(self, key):
        del self._store[key.lower()]

    def __repr__(self):
        return '%s(%r)' % (self.__class__.__name__, self._store)
    

class CaseInsensitiveDictEncoder(json.JSONEncoder):
    # might not need this
    def default(self, obj):
        if isinstance(obj, CaseInsensitiveDict):
            return obj.to_dict()
        # Let the base class default method raise the TypeError
        return super().default(obj)



"""Elements"""

# either

# self.add_element(Element(...))
# self.update_element("element_name", *args, **kwargs)

# or

# self.some_element = Element(...)
# self.add_element(element)
# self.some_element.update(*args, **kwargs)

"""Example Element"""
# old:
# doover_ui_variable(
#     "gaugePressure",
#     "Gauge Pressure (??)",
#     "float",
#     dec_precision=0,
#     ranges=[
#         {"label": "Low", "min": 0, "max": 5, "colour": "blue", "showOnGraph": True},
#         {"label": "Atmospheric", "min": 5, "max": 6, "colour": "green", "showOnGraph": True},
#         {"label": "High", "min": 5, "max": 10, "colour": "yellow", "showOnGraph": True},
#     ],
# ),

# new:
# ui.Variable(
#     "gaugePressure", "Gauge Pressure (??)", "float", dec_precision=0, ranges=[
#         Range("Low", min=0, max=5, colour=ui.Colour.blue),
#         Range("Atmospheric", min=5, max=6, colour=ui.Colour.green),
#         Range("High", min=6, max=10, colour=ui.Colour.yellow),
#     ]
# )

# or

# ui.NumericVariable(
#     "gaugePressure", "Gauge Pressure (??)", precision=0, ranges=[
#         Range("Low", min=0, max=5, colour=ui.Colour.blue),
#         Range("Atmospheric", min=5, max=6, colour=ui.Colour.green),
#         Range("High", min=6, max=10, colour=ui.Colour.yellow),
#     ]
# )

# other options could be BooleanVariable, TextVariable


"""Submodules"""

# either
class MyOtherSubmodule(...): ...

class MySubmodule(...):
    def __init__(self):
        self.add_element(...)
        self.add_element(...)
        self.add_submodule(MyOtherSubmodule())

# self.add_submodule(MySubmodule())

# or
# self.add_submodule(Submodule(Element(...), Element(...), Element(...))


"""COMMANDS"""
# note command callback only compatible with DDA/websockets
# either

class MyCommand(...):
    description = "..."
    def callback(self, *args, **kwargs):
        ...

# self.add_command(MyCommand())

# or

# @ui.command(description=...)
# def my_command(self, *args, **kwargs):
#     ...  # callback here



"""Critical Events"""

# either

# self.update_element("element_name", "value", critical=True)

# or define critical globally

# self.add_element(Element(..., critical=True))
# self.update_element("element_name", "value")


"""Log Manager"""

# self.set_log_interval(...)


"""Updating elements"""

# either

# self.update_element(...)
# self.push()  # alternatively, maybe self.sync() does a push and a pull? Should be in sync always, though...

# or

# self.update_element(..., immediate=True)  # maybe delayed=False?

# or

# with self.batch_update():
#      self.update_element(...)
#      self.update_element(...)
# implicit .push() is called


"""Coercing commands"""

# either

# self.my_command = MyCommand()
# self.add_command(my_command)
# somewhere else...
# self.my_command.coerce("some_value")

# or
# self.add_command(MyCommand())
# ...
# self.coerce_command("command_name", "some_value")

# or
# @ui.command(...)
# def my_command(...): ...
# ...
# self.my_command.coerce("some_value")


"""Significant Alerts / Crashes"""

# looks like this is just publishing to a channel. Can we just do

# self.api.publish_to_channel("significantAlerts", "some value")
# or
# self.dda_iface.publish_to_channel("significantAlerts", "some value")


ui1 = {'state': {'type': 'uiContainer', 'displayString': '', 'children': {'overviewPlot': {'type': 'uiMultiPlot', 'name': 'overviewPlot', 'displayString': 'Overview', 'series': ['lastTipTemp', 'lastTipHumidity', 'lastTipEmc'], 'colours': ['tomato', 'blue', 'limegreen'], 'activeSeries': [True, False, True]}, 'significantEvent': {'type': 'uiAlertStream', 'name': 'significantEvent', 'displayString': 'Notify me of any problems'}, 'selectedCrop': {'type': 'uiStateCommand', 'name': 'selectedCrop', 'displayString': 'Selected Crop', 'verboseString': 'Selected Crop', 'userOptions': {'spring_wheat': {'type': 'uiElement', 'name': 'spring_wheat', 'displayString': 'Spring Wheat'}, 'winter_wheat': {'type': 'uiElement', 'name': 'winter_wheat', 'displayString': 'Winter Wheat'}, 'durum_wheat': {'type': 'uiElement', 'name': 'durum_wheat', 'displayString': 'Durum Wheat'}, 'barley': {'type': 'uiElement', 'name': 'barley', 'displayString': 'Barley'}, 'peas': {'type': 'uiElement', 'name': 'peas', 'displayString': 'Peas'}, 'oats': {'type': 'uiElement', 'name': 'oats', 'displayString': 'Oats'}, 'canola': {'type': 'uiElement', 'name': 'canola', 'displayString': 'Canola'}, 'lentils': {'type': 'uiElement', 'name': 'lentils', 'displayString': 'Lentils'}, 'red_lentils': {'type': 'uiElement', 'name': 'red_lentils', 'displayString': 'Red Lentils'}, 'soybeans': {'type': 'uiElement', 'name': 'soybeans', 'displayString': 'Soybeans'}, 'flax': {'type': 'uiElement', 'name': 'flax', 'displayString': 'Flax'}, 'sunflower_seeds': {'type': 'uiElement', 'name': 'sunflower_seeds', 'displayString': 'Sunflower Seeds'}, 'long_grain_rice': {'type': 'uiElement', 'name': 'long_grain_rice', 'displayString': 'Long Grain Rice'}, 'corn': {'type': 'uiElement', 'name': 'corn', 'displayString': 'Corn'}, 'sorghum': {'type': 'uiElement', 'name': 'sorghum', 'displayString': 'Sorghum'}}, 'currentValue': 'oats'}, 'lastTipTemp': {'type': 'uiVariable', 'name': 'lastTipTemp', 'displayString': 'Tip Temperature (C)', 'form': 'radialGauge', 'varType': 'float', 'decPrecision': 1, 'ranges': [{'label': 'Low', 'min': 0, 'max': 20, 'colour': 'green', 'showOnGraph': False}, {'label': 'Warm', 'min': 20, 'max': 30, 'colour': 'yellow', 'showOnGraph': False}, {'label': 'High', 'min': 30, 'max': 40, 'colour': 'red', 'showOnGraph': False}], 'currentValue': -1}, 'lastTipHumidity': {'type': 'uiVariable', 'name': 'lastTipHumidity', 'displayString': 'Tip Humidity (%RH)', 'varType': 'float', 'decPrecision': 1, 'ranges': [{'label': 'Dry', 'min': 0, 'max': 30, 'colour': 'green', 'showOnGraph': False}, {'min': 30, 'max': 60, 'colour': 'blue', 'showOnGraph': False}, {'label': 'Humid', 'min': 60, 'max': 100, 'colour': 'yellow', 'showOnGraph': False}], 'currentValue': -1}, 'lastTipEmc': {'type': 'uiVariable', 'name': 'lastTipEmc', 'displayString': 'Tip Crop Moisture (%)', 'varType': 'float', 'decPrecision': 1, 'ranges': [{'min': 5, 'max': 8, 'colour': 'yellow', 'showOnGraph': False}, {'min': 8, 'max': 13, 'colour': 'blue', 'showOnGraph': False}, {'min': 13, 'max': 20, 'colour': 'yellow', 'showOnGraph': False}], 'currentValue': -1.0}, 'battVoltage': {'type': 'uiVariable', 'varType': 'float', 'name': 'battVoltage', 'displayString': 'Battery (V)', 'decPrecision': 1, 'ranges': [{'label': 'Low', 'min': 5.0, 'max': 6.0, 'colour': 'yellow', 'showOnGraph': True}, {'min': 6.0, 'max': 6.5, 'colour': 'blue', 'showOnGraph': True}, {'label': 'Good', 'min': 6.5, 'max': 7.5, 'colour': 'green', 'showOnGraph': True}, {'label': 'Over Voltage', 'min': 7.5, 'max': 8.0, 'colour': 'yellow', 'showOnGraph': True}], 'currentValue': 7.140415}, 'alarm_settings_submodule': {'type': 'uiSubmodule', 'name': 'alarm_settings_submodule', 'displayString': 'Alarms', 'children': {'emcAlarmSlider': {'type': 'uiSlider', 'name': 'emcAlarmSlider', 'displayString': 'Moisture Alarm (%)', 'min': 5, 'max': 30, 'stepSize': 0.1, 'dualSlider': True, 'icon': 'fa-regular fa-bell', 'showActivity': False, 'isInverted': False}, 'humidityAlarmSlider': {'type': 'uiSlider', 'name': 'humidityAlarmSlider', 'displayString': 'Relative Humidity Alarm (%)', 'min': 30, 'max': 95, 'stepSize': 0.1, 'dualSlider': True, 'showActivity': False, 'isInverted': False, 'icon': 'fa-regular fa-bell'}, 'tempAlarmSlider': {'type': 'uiSlider', 'name': 'tempAlarmSlider', 'displayString': 'Temperature Alarm (˚C)', 'min': -10, 'max': 50, 'stepSize': 0.1, 'showActivity': False, 'dualSlider': True, 'isInverted': False, 'icon': 'fa-regular fa-bell'}, 'batteryAlarm': {'type': 'uiSlider', 'name': 'batteryAlarm', 'displayString': 'Battery Alarm (V)', 'showActivity': False, 'min': 5, 'max': 10, 'stepSize': 0.1, 'dualSlider': False, 'isInverted': False}}}, 'details_submodule': {'type': 'uiSubmodule', 'name': 'details_submodule', 'displayString': 'Details', 'children': {'sleepTime': {'type': 'uiFloatParam', 'name': 'sleepTime', 'displayString': 'Periodic Sleep Time (mins)', 'min': 5, 'max': 1440}, 'pumpTime': {'type': 'uiFloatParam', 'name': 'pumpTime', 'displayString': 'Pump Time (secs)', 'min': 0, 'max': 300}, 'lastHeadTemp': {'type': 'uiVariable', 'name': 'lastHeadTemp', 'displayString': 'Head Temperature (C)', 'form': 'radialGauge', 'varType': 'float', 'decPrecision': 1, 'ranges': [{'label': 'Low', 'min': 0, 'max': 20, 'colour': 'green', 'showOnGraph': False}, {'label': 'Warm', 'min': 20, 'max': 30, 'colour': 'yellow', 'showOnGraph': False}, {'label': 'High', 'min': 30, 'max': 40, 'colour': 'red', 'showOnGraph': False}], 'currentValue': 16.07042}, 'lastHeadHumidity': {'type': 'uiVariable', 'name': 'lastHeadHumidity', 'displayString': 'Head Humidity (%RH)', 'varType': 'float', 'decPrecision': 1, 'ranges': [{'label': 'Dry', 'min': 0, 'max': 30, 'colour': 'green', 'showOnGraph': False}, {'min': 30, 'max': 60, 'colour': 'blue', 'showOnGraph': False}, {'label': 'Humid', 'min': 60, 'max': 100, 'colour': 'yellow', 'showOnGraph': False}], 'currentValue': 71.0428}}}, 'dashboard_submodule': {'type': 'uiSubmodule', 'name': 'dashboard_submodule', 'displayString': 'Dashboard Details', 'children': {'capacity': {'type': 'uiFloatParam', 'name': 'capacity', 'displayString': 'Maximum Storage Capacity (T)', 'min': 0, 'max': 10000, 'currentValue': 2000}, 'currentVolume': {'type': 'uiFloatParam', 'name': 'currentVolume', 'displayString': 'Current Volume (T)', 'min': 0, 'max': 10000, 'currentValue': 1800}}}, 'node_connection_info': {'type': 'uiConnectionInfo', 'name': 'node_connection_info', 'connectionType': 'periodic', 'connectionPeriod': 21600, 'nextConnection': 21600, 'allowedMisses': 4}}}}
ui2 = {'state': {'type': 'uiContainer', 'display_str': 'Device IO Tester', 'show_activity': True, 'children': {'disubmodule': {'name': 'disubmodule', 'type': 'uiSubmodule', 'display_str': 'Digital Inputs', 'show_activity': True, 'children': {'di-fetchnow': {'name': 'di-fetchnow', 'type': 'uiAction', 'display_str': 'Fetch Status Now', 'show_activity': True, 'colour': 'blue', 'requiresConfirm': True}, 'di-warning-1': {'name': 'di-warning-1', 'type': 'uiWarningIndicator', 'display_str': 'Digital Inputs are Scary!', 'show_activity': True, 'can_cancel': True}, 'di0-state': {'name': 'di0-state', 'type': 'uiVariable', 'display_str': 'DI0 State', 'show_activity': True, 'varType': 'bool', 'ranges': []}, 'di1-state': {'name': 'di1-state', 'type': 'uiVariable', 'display_str': 'DI1 State', 'show_activity': True, 'varType': 'bool', 'ranges': []}}}, 'dosubmodule': {'name': 'dosubmodule', 'type': 'uiSubmodule', 'display_str': 'Digital Outputs', 'show_activity': True, 'children': {'do0-toggle': {'name': 'do0-toggle', 'type': 'uiFloatParam', 'display_str': 'Digital Output 0 Toggle', 'show_activity': True}, 'do1-toggle': {'name': 'do1-toggle', 'type': 'uiFloatParam', 'display_str': 'Digital Output 1 Toggle', 'show_activity': True}, 'do0-state': {'name': 'do0-state', 'type': 'uiVariable', 'display_str': 'Digital Output {i} State', 'show_activity': True, 'varType': 'bool', 'ranges': []}, 'do1-state': {'name': 'do1-state', 'type': 'uiVariable', 'display_str': 'Digital Output {i} State', 'show_activity': True, 'varType': 'bool', 'ranges': []}, 'do2-state': {'name': 'do2-state', 'type': 'uiVariable', 'display_str': 'Digital Output {i} State', 'show_activity': True, 'varType': 'bool', 'ranges': []}, 'do3-state': {'name': 'do3-state', 'type': 'uiVariable', 'display_str': 'Digital Output {i} State', 'show_activity': True, 'varType': 'bool', 'ranges': []}, 'do4-state': {'name': 'do4-state', 'type': 'uiVariable', 'display_str': 'Digital Output {i} State', 'show_activity': True, 'varType': 'bool', 'ranges': []}, 'do2-toggle': {'name': 'do2-toggle', 'type': 'uiFloatParam', 'display_str': 'Digital Output 2 Toggle', 'show_activity': True}, 'do3-toggle': {'name': 'do3-toggle', 'type': 'uiFloatParam', 'display_str': 'Digital Output 3 Toggle', 'show_activity': True}, 'do4-toggle': {'name': 'do4-toggle', 'type': 'uiFloatParam', 'display_str': 'Digital Output 4 Toggle', 'show_activity': True}}}, 'ai-submodule': {'name': 'ai-submodule', 'type': 'uiSubmodule', 'display_str': 'Analog Inputs', 'show_activity': True, 'children': {'num-ai-var': {'name': 'num-ai-var', 'type': 'uiFloatParam', 'display_str': 'Number of Analog Inputs', 'show_activity': True}, 'ai-0-state': {'name': 'ai-0-state', 'type': 'uiVariable', 'display_str': 'Analog Input 0', 'show_activity': True, 'varType': 'float', 'ranges': []}, 'ai-1-state': {'name': 'ai-1-state', 'type': 'uiVariable', 'display_str': 'Analog Input 1', 'show_activity': True, 'varType': 'float', 'ranges': []}, 'ai-2-state': {'name': 'ai-2-state', 'type': 'uiVariable', 'display_str': 'Analog Input 2', 'show_activity': True, 'varType': 'float', 'ranges': []}, 'ai-3-state': {'name': 'ai-3-state', 'type': 'uiVariable', 'display_str': 'Analog Input 3', 'show_activity': True, 'varType': 'float', 'ranges': []}}}, 'sys-temp': {'name': 'sys-temp', 'type': 'uiVariable', 'display_str': 'System Temp', 'show_activity': True, 'varType': 'float', 'ranges': []}, 'sys-voltage': {'name': 'sys-voltage', 'type': 'uiVariable', 'display_str': 'System Voltage', 'show_activity': True, 'varType': 'float', 'ranges': []}, 'online': {'name': 'online', 'type': 'uiVariable', 'display_str': 'Online', 'show_activity': True, 'varType': 'time', 'currentValue': 1717025334.87972, 'ranges': []}}}}



import json
import ast
import re

def processor_name(name: str) -> str:
    return "#" + name.lstrip("#")


def task_name(name: str) -> str:
    return "!" + name.lstrip("!")


def maybe_json(data: str):
    try:
        return json.loads(data)
    except json.JSONDecodeError:
        pass

    num_occ = len([n for n in data if n == "'"])
    if num_occ > 0 and num_occ % 2 == 0 and len([n for n in data if n == "{"]) == len([n for n in data if n == "}"]):
        return maybe_json(data.replace("'", '"'))

    return data

def int_or_list(data: str):
    data = data.strip()
    if data.startswith("[") and data.endswith("]"):
        data_list = ast.literal_eval(data)
        for element in data_list:
            if not isinstance(element, int):
                raise TypeError(f"Inputted {data} is not a list of int or a single int")
        return data_list
    else:
        return int(data)
    
def bool_or_list(data: str):
    data = data.strip()
    if data.startswith("[") and data.endswith("]"):
        data_list = ast.literal_eval(data)
        for element in data_list:
            if not isinstance(element, bool):
                raise TypeError(f"Inputted {data} is not a list of int or a single int")
        return data_list
    else:
        try:
            data_eval = ast.literal_eval(data)
        except:
            raise TypeError(f"Inputted {data} is not a list of int or a single int")
        if not isinstance(data_eval, bool):
            raise TypeError(f"Inputted {data} is not a list of int or a single int")
        return data_eval
    
def float_or_list(data: str):
    data = data.strip()
    if data.startswith("[") and data.endswith("]"):
        data_list = ast.literal_eval(data)
        for element in data_list:
            if not isinstance(element, float):
                raise TypeError(f"Inputted {data} is not a list of int or a single int")
        return data_list
    else:
        return float(data)
    
def json_or_str(data: str):
    try:
        return json.dumps(data)
    except:
        return str

def extract_parameters(docstring: str):
    """Extract parameters from a NumPy-style docstring as a list of tuples (name, type, description)."""
    pattern = r"Parameters\n-{10,}\n(.*?)(?=\n\S|\Z)"  # Match 'Parameters' section
    match = re.search(pattern, docstring, re.DOTALL)
    
    if not match:
        return {}

    params_text = match.group(1).strip()
    param_lines = params_text.split("\n")

    params = {}
    param_name, param_type, param_desc = None, None, []
    
    for line in param_lines:
        line = line.strip()
        if ": " in line:  # This line defines a parameter name and type
            if param_name:  # Store the previous parameter
                params[param_name] = (param_type, " ".join(param_desc).strip())

            param_name, param_type = map(str.strip, line.split(": ", 1))
            param_desc = []  # Reset description buffer
        else:
            param_desc.append(line)  # Append description lines

    if param_name:  # Store the last parameter
        params[param_name] = (param_type, " ".join(param_desc).strip())

    return params

def extract_description(docstring: str) -> str:
    """Extract everything before the 'Parameters' section in a NumPy-style docstring."""
    pattern = r"^(.*?)(?=\nParameters\n-{10,})"  # Capture everything before 'Parameters'
    match = re.search(pattern, docstring, re.DOTALL)
    return match.group(1).strip() if match else docstring.strip()


class BoolFlag:
    def __call__(self, *args, **kwargs):
        return



import asyncio
import logging
import time
import typing

from .camera import CameraInterface
from ...ui import Camera as UICamera, SlimCommand, RemoteComponent, Submodule
from ...ui import CameraHistory as UICameraHistory

if typing.TYPE_CHECKING:
    from ...utils import CaseInsensitiveDict


class Camera:
    def __init__(self, config: "CaseInsensitiveDict[str, str]"):
        self.config = config

        self.name = config.get("name", "camera").lower().replace(' ', '_')
        self.display_name = config.get("display_name") or self.name
        self.uri = config.get("uri")
        self.mode = config.get("mode", "mp4")
        self.snapshot_secs = config.get("snapshot_secs", 6)
        self.wake_delay = config.get("wake_delay", 10)

        self.remote_component_url = config.get("remote_component_url")
        self.remote_component_name = config.get("remote_component_name")
        self.address = config.get("address")
        self.port = config.get("rtsp_port")
        self.cam_type = config.get("type")


    def __eq__(self, other):
        return other is not None and isinstance(other, self.__class__) and self.config == other.config

    async def trigger_snapshot(self, cam_iface: "CameraInterface"):
        return await cam_iface.get_camera_snapshot_async(self.uri, self.name, self.mode, self.snapshot_secs)


## CONFIG WILL LOOK LIKE THE FOLLOWING
#
# {
#     "camera_config": {
#         "power_pin" : 1,
#         "cameras": {
#             "cam_1" : {
#                 "name": "cam_1",
#                 "display_name": "Camera 1",
#                 "uri": "rtsp://192.168.50.120:544",
#                 "mode": "mp4",
#                 "snapshot_secs": 6
#             },
#             "cam_2" : {

#             }
#         }
#     }
# }


class CameraManager:
    def __init__(self,
                camera_config=None,
                config_manager=None,
                dda_iface=None,
                ui_manager=None,
                plt_iface=None,
                app_manager=None,
                is_active=True,
            ):
        self.camera_config = camera_config
        self.config_manager = config_manager
        self.dda_iface = dda_iface
        self.ui_manager = ui_manager
        self.plt_iface = plt_iface
        self.app_manager = app_manager

        self._is_active = is_active

        self.last_camera_snapshot = time.time()
        self.iter_counter = 0
        self.snapshot_task = None

        self.camera_snap_cmd_name = "camera_snapshots"
        self.last_snapshot_cmd_name = "last_cam_snapshot"

        self.cameras: list[Camera] = []

    async def load_camera_config(self):
        if self.config_manager is not None:
            self.camera_config = await self.config_manager.get_config_async("camera_config")

    def load_cameras(self):
        try:
            if not self.camera_config:
                logging.warning("No cameras found in config")
                return
            self.cameras = [Camera(cam_config) for cam_config in self.camera_config["CAMERAS"].values()]
        except (KeyError, AttributeError):
            logging.warning("Unable to parse cameras from config")

    @property
    def is_active(self):
        return self._is_active and self.camera_config is not None

    def get_is_active(self):
        return self.is_active

    def get_camera_snapshot_period(self):
        if not self.is_active:
            return None

        try:
            return self.camera_config["SNAPSHOT_PERIOD"]
        except KeyError:
            logging.warning("Unable to parse camera snapshot period from config")
        return None

    def fetch_ui_elements(self):
        if not self.is_active:
            return []

        components = []
        for cam in self.cameras:
            if cam.remote_component_url is not None:

                liveview_element_name = f"{cam.name}_liveview"
                liveview_display_name = f"{cam.display_name} Liveview"
                ui_liveview = RemoteComponent(
                    name=liveview_element_name,
                    display_name=liveview_display_name,
                    cam_name=cam.name,
                    component_url=cam.remote_component_url,
                    address=cam.address,
                    port=cam.port,
                    rtsp_uri=cam.uri,
                    cam_type=cam.cam_type,
                )
                # pretty hacky, but this basically tells the UI to never overwrite these fields since
                # we manage them in the camera interface. Possibly not the right way of going about it?
                ui_liveview._retain_fields = ("presets", "active_preset", "cam_position", "allow_absolute_position")

                ## Set the Dispaly Name to blank to avoid title in submodule
                original_cam_history = UICameraHistory(cam.name, "", cam.uri)

                components.append(
                    UICamera(cam.name, cam.display_name, cam.uri,
                        children=[
                            ui_liveview,
                            original_cam_history
                        ]
                ))

            else:
                components.append(UICamera(cam.name, cam.display_name, cam.uri))

        return components

    def update_ui_elements(self):
        pass

    async def setup(self):
        await self.load_camera_config()
        self.load_cameras()

        self.ui_manager.add_children(*self.fetch_ui_elements())
        self.ui_manager._add_interaction(SlimCommand(self.camera_snap_cmd_name, callback=self.on_snapshot_command))

    async def main_loop(self):
        self.iter_counter += 1
        if self.iter_counter > 999999999:
            self.iter_counter = 0

        if self.iter_counter % 10 == 0: ## Only check if cameras overdue on every 10th cycle
            self.assess_snapshot_due()

    def assess_snapshot_due(self):
        if not self.get_is_active():
            return False
        if self.snapshot_task is not None:
            return False ## Snapshot currently running
        snap_period = self.get_camera_snapshot_period()
        if snap_period is None:
            return False
        if time.time() - self.last_camera_snapshot > snap_period:
            self.trigger_all_snapshots()
            return True
        return False

    async def on_snapshot_command(self, new_value: str):
        if new_value != "get_immediate_snapshot":
            return

        self.ui_manager.coerce_command(self.camera_snap_cmd_name, None)
        if self.snapshot_task is None:
            self.trigger_all_snapshots()

    def set_last_snapshot_time(self, ts=None):
        ts = ts or time.time()

        self.last_camera_snapshot = ts
        self.ui_manager.coerce_command(self.last_snapshot_cmd_name, ts)

    def trigger_all_snapshots(self):
        if self.snapshot_task is not None:
            logging.info("Skipping trigger snapshot request, snapshot task already running")
            return

        self.set_last_snapshot_time()
        self.snapshot_task = asyncio.create_task(self.run_all_snapshots_task())
        self.snapshot_task.add_done_callback(self._on_snapshot_done)

    def _on_snapshot_done(self, task: asyncio.Task) -> None:
        try:
            exc = task.exception()
        except asyncio.CancelledError:
            exc = None

        if exc:
            logging.error(f"Snapshot task failed: {exc}", exc_info=exc)

        self.snapshot_task = None

    async def run_all_snapshots_task(self):
        if len(self.cameras) < 1:
            logging.warning("No cameras found for snapshot task.")
            return

        for c in self.cameras:
            await c.trigger_snapshot(self.app_manager.get_camera_iface())

        logging.info("Camera take all snapshots complete")

    async def is_snapshot_running(self):
        return await self.app_manager.get_camera_iface().is_snapshot_running_async()


camera_manager = CameraManager
camera = Camera


import base64
import os
import shutil
import uuid
import mimetypes
import logging
import sys
import importlib
import pathlib
from datetime import datetime

from typing import TYPE_CHECKING, Any, Optional

if TYPE_CHECKING:
    from .client import Client


class Channel:

    def __init__(self, *, client, data):
        self.client: "Client" = client
        self._aggregate = None
        self._messages = None

        self._from_data(data)

    def __eq__(self, other):
        return isinstance(other, self.__class__) and self.id == other.id

    def _from_data(self, data):
        self.id = data["channel"]
        self.name = data["name"]
        # from the get_agent endpoint this is `agent`, from the get_channel endpoint this is `owner`.
        self.agent_id = (data.get("owner") or data.get("agent"))
        self._agent = None

        try:
            self._aggregate = data["aggregate"]["payload"]
        except KeyError:
            self._aggregate = None

    @property
    def aggregate(self):
        return self._aggregate

    def update(self):
        res = self.client._get_channel_raw(self.id)
        self._from_data(res)

    def get_tunnel_url(self, address):
        if self.name != "tunnels":
            raise RuntimeError("Tunnels are only valid in the `tunnels` channel.")

        agg = self.fetch_aggregate()
        try:
            tunnels = agg["open"]
        except KeyError:
            return

        found = [t for t in tunnels if t["address"] == address]
        if found:
            return found[0]["url"]

    def fetch_agent(self):
        if self._agent is not None:
            return self._agent

        self._agent = self.client.get_agent(self.agent_id)
        return self._agent

    def fetch_aggregate(self):
        if self._aggregate is not None:
            return self._aggregate
    
        self.update()
        return self._aggregate

    def fetch_messages(self, num_messages: int = 10):
        if self._messages is not None:
            return self._messages

        self._messages = self.client.get_channel_messages(self.id, num_messages=num_messages)
        return self._messages

    def fetch_messages_in_window(self, window_start: datetime, window_end: datetime):
        return self.client.get_channel_messages_in_window(self.id, window_start, window_end)

    def publish(self, data: Any, save_log: bool = True, log_aggregate: bool = False, override_aggregate: bool = False, timestamp: Optional[datetime] = None):
        return self.client.publish_to_channel(self.id, data, save_log, log_aggregate, override_aggregate, timestamp)

    @property
    def last_message(self):
        messages = self.fetch_messages(num_messages=1)
        if messages is None or len(messages) == 0:
            return None
        return messages[0]
    
    @property
    def last_update_age(self):
        last_message = self.last_message
        if last_message is None:
            return None
        return last_message.get_age()

    def update_from_file(self, file_path, mime_type=None):
        if mime_type is None:
            mime_type, _ = mimetypes.guess_type(file_path)
            # mime_type = "application/octet-stream"

        with open(file_path, "rb") as f:
            b64_data = base64.b64encode(f.read()).decode()

        msg = {
            "output_type": mime_type,
            "output": b64_data
        }
        self.publish(msg)


class Processor(Channel):

    def update_from_package(self, package_dir):
        fp = f"/tmp/{uuid.uuid4()}"
        shutil.make_archive(fp, 'zip', package_dir)

        with open(f"{fp}.zip", "rb") as f:
            zip_bytes = f.read()
            b64_package = base64.b64encode(zip_bytes).decode()

        self.publish(b64_package)
        os.remove(f"{fp}.zip")

    def invoke_locally(self, 
            package_dir,
            agent_id,
            access_token,
            api_endpoint="https://my.doover.dev",
            package_config={},
            msg_obj={},
            task_id=None,
            log_channel=None,
            agent_settings={},
            # *args, **kwargs
        ):
        
        logging.basicConfig(level=logging.DEBUG)
        sys.path.append(package_dir)

        # Construct the full path to "target.py" within package_dir
        target_path = os.path.join(package_dir, "target.py")

        ## import the loaded generator file
        spec = importlib.util.spec_from_file_location("target", target_path)
        target_task = importlib.util.module_from_spec(spec)
        spec.loader.exec_module(target_task)

        # from .target import generator
        target_task = getattr(target_task, 'target')

        #     'agent_id' : The Doover agent id invoking the task e.g. '9843b273-6580-4520-bdb0-0afb7bfec049'
        #     'access_token' : A temporary token that can be used to interact with the Doover API .e.g 'ABCDEFGHJKLMNOPQRSTUVWXYZ123456890',
        #     'api_endpoint' : The API endpoint to interact with e.g. "https://my.doover.com",
        #     'package_config' : A dictionary object with configuration for the task - as stored in the task channel in Doover,
        #     'msg_obj' : A dictionary object of the msg that has invoked this task,
        #     'task_id' : The identifier string of the task channel used to run this processor,
        #     'log_channel' : The identifier string of the channel to publish any logs to
        #     'agent_settings' : {
        #       'deployment_config' : {} # a dictionary of the deployment config for this agent
        task_obj = target_task(
            agent_id=agent_id,
            access_token=access_token,
            api_endpoint=api_endpoint,
            package_config=package_config,
            msg_obj=msg_obj,
            task_id=task_id,
            log_channel=log_channel,
            agent_settings=agent_settings,
            # *args, **kwargs,
        )

        task_obj.execute()


class Task(Channel):

    def _from_data(self, data):
        super()._from_data(data)
        self.processor_id: str = data.get("processor")
        self._processor = None

    def fetch_processor(self) -> Optional[Processor]:
        if self._processor is not None:
            return self._processor
        if self.processor_id is None:
            return

        self._processor = self.client.get_channel(self.processor_id)
        return self._processor

    def subscribe_to_channel(self, channel_id: str):
        return self.client.subscribe_to_channel(channel_id, self.id)

    def unsubscribe_from_channel(self, channel_id: str):
        return self.client.unsubscribe_from_channel(channel_id, self.id)

    def invoke_locally(self, package_dir, msg_obj, agent_settings, agent_id=None):
        processor = self.fetch_processor()
        if processor is None:
            return
        
        agent_id = agent_id or self.client.agent_id
        access_token = self.client.access_token.token
        api_endpoint = self.client.base_url
        package_config = self.fetch_aggregate()
        task_id = self.id

        log_channel = None

        processor.invoke_locally(
            package_dir,
            agent_id,
            access_token,
            api_endpoint,
            package_config,
            msg_obj,
            task_id,
            log_channel,
            agent_settings,
        )